#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "KissBLE.h"
#include "KissTimer.h"
#include "BitMapCache.h"
#include "VirtualPool.h"
#include "kissRedefines.h"
#include "bitmapUtils.h"
#include "mbed.h"
//#include "AppLog.h"
//#include "ResponseStream.h"

#if defined(APP_LOG)
#include "KissLogger.h"
#else
#include "emptyKissLogger.h"
#endif
#if defined(KISS_EMULATOR)
#include "KissEmulator.h"
#else
#include "emptyKissEmulator.h"
#endif

#if !defined(EQUEUE_WAIT_MAX)
#error "No wait time safeguard in equeue module"
#endif

//extern ResponseStream	*rsOut;

const short KissBLE::channels[KissBLE::MAX_CHANNELS + 1] = {-1, 1, 2, 3};
const NODE_ID KissBLE::nullNodeId = {0};

#define COMMA			,
#define VSIZE(a) a COMMA sizeof(a)

#define BUILD_BUG_ON(condition)			((void)sizeof(char[1 - 2*!!(condition)]))
#define ROUTER_INACTIVE_MAX				4
#define CHECK_PHASE_MINIMAL_INTERVAL	500		// microseconds
#define	TRANSMIT_DELAY_US				300		// uS
#define	MIN_TRANSMIT_DELAY_US			150		// uS

#define CHANNEL(idx)		(KissBLE::channels[idx])
#define TICK_TIME			(activeSchema.tickTime * KissBLE::TICKS_SCALE)
#define TICK_TIME_US		(TICK_TIME * 1000)
#define NODE_EQ(n1,n2)		(nodeEQ(n1, n2))
#define CONN_EQ(n1,n2)		(connEQ(n1, n2))
#define THIS_NODE(nd)		(NODE_EQ(nd, netParams.idParams.node))
#define IS_DEVICE			(thisRouterId == 0)
#define EMPTY_NODE(a)		(KissBLE::nodeIdNull(a))
#define IS_REGISTERED		(!EMPTY_NODE(netParams.idParams.node))

#define	INTERVAL_TIME		(TICK_TIME * (activeSchema.maxTickCount + 1))
#define USER_FRAME(ft)		(ft <= KissBLE::CTRL_FRAME_TYPE_BASE && ft > 0)
#define CTRL_FRAME(ft)		(ft > KissBLE::CTRL_FRAME_TYPE_BASE)
#define SCHEMA_ACTIVE		(activeSchema.tickTime > 0)
#define TICK_SHIFT_US(sft)	(sft * KissBLE::SHIFT_SCALE)
#define SPECIAL_ROLE(nd)	(nd.role == KissBLE::BORDER_ROUTER || nd.role == KissBLE::SYNC_ROUTER)
#define RESERVED_ROUTER		(netParams.idParams.reserveRole == KissBLE::BORDER_ROUTER || \
							netParams.idParams.reserveRole == KissBLE::SYNC_ROUTER)
#define INIT_CHANNEL_IDX	(RESERVED_ROUTER ? KissBLE::SYNC_ROUTER_CHANNEL_IDX : (syncLoop % KissBLE::MAX_CHANNELS) + 1)
// Advertisement frame received lately (because schema updated!)
#define ADVERT_ACTIVE		(tickStartTime == schemaSyncTime)
#define FREE_COMMAND(idx, ...)	freeCommand(idx, __LINE__, ##__VA_ARGS__)
//#define ISR_IS_ACTIVE		core_util_is_isr_active()

#define LOG_LEVEL_MINIMUM		KissLogger::LOG_LEVEL_MINIMUM
#define LOG_LEVEL_FRAMES		KissLogger::LOG_LEVEL_FRAMES
#define LOG_LEVEL_DETAILS		KissLogger::LOG_LEVEL_DETAILS
#define LOG_LEVEL_ALL_FRAMES	KissLogger::LOG_LEVEL_ALL_FRAMES
#define LOG_LEVEL_FULL			KissLogger::LOG_LEVEL_FULL
#define ROUTER_TIME				KissTimer::routerTime

static IRQn_Type			radioInterruptNo;

// ------------------ Global variables -----------------------
EventQueue				*kissQueue;
static Transceiver		*radio;
int16_t					framesInQueue;		// number of frames waiting for process
//------------------ advertising variables -----------------------
static uint8_t			frameSequence;

// ------------------ Current phase variables ----------------------
uint32_t				currentIntervalNo;		// current interval number
PROCESS_PHASE			currentPhase;			// process state
TIME_SLOT_MODE			slotModeTab[1 << KissBLE::MAX_TICKS_BITS];
short					currentSlot;
uint64_t				currentSlotStart;
uint64_t				currentSlotEnd;
short					currentSlotChannel;
// Time of last frame sended or time of received frame with EOT bit set.
static volatile uint64_t	currentSlotCompletedTime;
static uint32_t			frames2sendMap;

// ----------------- Schema synchronisation variables ---------------
NODE_ID						masterRouterNode;
static CTRL_RADVERT_BODY	routerAdvertisement;
static uint32_t				eosFromAdvertisement;
COMM_SCHEMA					activeSchema;
static uint64_t				netStartTime;
volatile uint64_t			tickStartTime;			// start time of activity interval
uint64_t					schemaSyncTime;			// tickStartTime from last adv frame
BitMapCache<NODE_STATE, KissBLE::MAX_LOCAL_NODES> 
							localNodes;
BitMapCache<CONNECTION_STATE, KissBLE::MAX_CONNECTIONS> 
							connections;
static int8_t				routerInactiveCount;
static int8_t				thisRouterId;

static uint16_t				sendTimeCorrector;		// for precise correction time in Advert frames
static uint16_t				timeCorrectorCount;

// Device transmit gaps (in extra & normal timeslot)
static Timeout				transmitTimeControl;
static Timeout				newPhaseTimeControl;

// ----------------- Init phase variables ---------------
static int8_t	syncLoop;			// current listen channel  in SYNC_PHASE

static short				(*appHandler)(EVENT *);
KissSecurityManager *securityManager;
short						securityMethod;
// Random generator for security manager 'salt' generation and random time send shift ('lantern')
short						(*randomData)(void *buf, uint8_t bufLen);

// ---------------- kissQueue for command  and response ------------------

BitMapCache<COMMAND_STATE, KissBLE::CMD_CACHE_SIZE> cmdCache;
uint8_t packetPool[32 * KissBLE::POOL_CHUNK_SIZE];
VirtualPool memPool(sizeof(packetPool), KissBLE::POOL_CHUNK_SIZE, packetPool);
static	uint32_t cmdPending4handler;

//------------------ Parameters ---------------------------
NETWORK_PARAMS	netParams;

//------------------- send delay estimation ---------------------
static volatile uint64_t	lastFrameSendStop;
static volatile uint64_t	lastFrameSendStart;
volatile short				estimatedFrameDelay;// statistical
static volatile uint8_t		delayMeasurments;

//------------------- forward declarations ------------------------
static void		processReceivedPacket(IFRAME ihead);
static short	callAppHandler(short event_type, short cmdIdx = -1);
static void		setCurrentPhase(FRAME *frame);
static void		syncPhase();
static void		radioHandler(short event);

/*
static short serializeFrame(FRAME *frame, uint8_t *out, uint16_t maxPacketSize)
{
	if(maxPacketSize > sizeof(FRAME))
		return -1;
	memcpy(out, frame, sizeof(FRAME));
	return KissBLE::getLength(frame);
}

static short deserializeFrame(FRAME *frame, uint8_t *in, uint16_t packetSize)
{
	if(packetSize > sizeof(FRAME))
		return -1;
	memcpy(frame, in, packetSize);
	return KissBLE::getLength(frame);
}
*/
bool connEQ(const NODE_ID id1, const NODE_ID id2)
{
	union {
		NODE_ID		node;
		uint16_t	whole;
	} u1, u2;
	BUILD_BUG_ON(sizeof(NODE_ID) != sizeof(uint16_t));
	u1.node = id1;
	u2.node = id2;
	return u1.whole == u2.whole;
}

bool nodeEQ(NODE_ID id1, NODE_ID id2)
{
	id1.connPort = id2.connPort = 0;
	return connEQ(id1, id2);
}

#if !defined(KISS_EMULATOR)	
static inline short getVirtualNode(MAC_ADDR *mac)
{
	return (KissBLE::macEQ(mac, &netParams.idParams.mac)) ? 0 : -1;
}

static inline short getVirtualNode(NODE_ID node)
{
	return (nodeEQ(node, netParams.idParams.node)) ? 0 : -1;
}

#endif	

NODE_STATE *getNodeState(short node)
{
	return localNodes.getCellContent(node);
}

inline NODE_STATE *setNodeState(NODE_STATE *buf, short idx = -1)
{
	return localNodes.put(buf, idx);
}

COMMAND_STATE *getCommandState(short cmdidx)
{
	return cmdCache.getCellContent(cmdidx);
}

short getNextCommand(COMMAND_STATE **ns)
{
	return cmdCache.nextBusyCell(ns);
}

COMMAND_STATE *setCommandState(COMMAND_STATE *buf, short idx = -1)
{
	return cmdCache.put(buf, idx);
}

FHEAD *getHeader(COMMAND_STATE *cs)
{
	FHEAD *rv = NULL;
	if(cs->codedLen || cs->omitted || cs->poolHandle == 0)
		rv = &cs->head;
	else
	{
		FRAME *frame = (FRAME *) memPool.getAddress(cs->poolHandle);
		if(frame)
			rv = &frame->shead;
	}
	return rv;
}

void freeCommand(short idx, short line, short place = -1)
{
	COMMAND_STATE *cs = cmdCache.getCellContent(idx);
	if(cs)
	{
		short stat = 0;
		if(cs->poolHandle)
			stat = memPool.free(cs->poolHandle);
		LOGGER(logFreeCommand, cs, line, place, stat);
		(void) stat;
		cs->poolHandle = 0;
		cmdCache.freeCell(idx);
	}
}

bool stopProcessing(bool clearQueue = true)
{
	radio->stopReceive(clearQueue);
	transmitTimeControl.detach();
	newPhaseTimeControl.detach();
	return true;
}

FRAME *getFrame(COMMAND_STATE *cs, bool notCoded = false)
{
	short stat, idx = cmdCache.getCellId(cs);
	FRAME *frame = (FRAME *) memPool.getAddress(cs->poolHandle);
	if(frame && cs->codedLen)
	{	// decode
		FRAME tmp = {0};
		if(notCoded)
			return NULL;
		stat = securityManager->decodeFrame((uint8_t *) frame, cs->codedLen, (uint8_t *) &tmp);
		if(stat > 0)
		{
			stat = KissBLE::getLength(&tmp);
			memcpy(frame, &tmp, stat);
			cs->codedLen = 0;
		}
		else
		{
			FREE_COMMAND(idx, 0);
			frame = NULL;
		}
	}
	return frame;
}

uint32_t composeEOS(FHEAD *head)
{
	uint16_t *nd = (head->response) ? (uint16_t *) &head->dstNode : (uint16_t *) &head->srcNode;
	uint32_t rv = (*nd) | (head->frameSeq << 16) | (head->frameType << 24) | (head->response << 28);
	return rv;
}

static inline bool checkEOS(FHEAD *head, uint32_t eosIdent)
{
	return (composeEOS(head) == eosIdent);
}

NODE_ID composeNodeId(short nodeIdx)
{
	NODE_ID rv = netParams.idParams.node;	// prototype
	if(nodeIdx == KissBLE::ROUTER_NODEID)
		return (IS_DEVICE) ? masterRouterNode : rv;
	rv.role = KissBLE::DEVICE;
	rv.nodeId = nodeIdx;
	return rv;
}

short getLocalNode(NODE_ID node, uint16_t *localMask = NULL)
{
	if(node.routerId == netParams.idParams.node.routerId)
	{
		short rv = node.nodeId;
		if(rv >= localNodes.CACHE_SIZE)
			return -1;
		if(rv > 0)
		{
			if(localMask)
				*localMask |= (0x8000 >> rv);
			return rv;
		}
		else if(rv == 0 && localMask)
			*localMask = localNodes.getBitmap() >> 16;
		return rv;
	}
	return -1;
}

void switch2routerTime(IFRAME *iframe)
{
	if(!iframe->timeSchema)
	{
		if(iframe->startTime)
			iframe->startTime = KissTimer::routerTime(iframe->startTime);
		if(iframe->endTime)
			iframe->endTime = KissTimer::routerTime(iframe->endTime);
		iframe->timeSchema = true;
	}
}

// calculate minimal node distance in hops
static short frameHoops(FHEAD *head)
{
	short rv;
	if(EMPTY_NODE(head->srcNode))
		rv = 2;	// frame dst is router (Register frame)
	else if(head->dstNode.routerId == head->srcNode.routerId || EMPTY_NODE(head->dstNode))
	{	// nodes on the same router
		if(getLocalNode(head->dstNode) == KissBLE::ROUTER_NODEID ||
			getLocalNode(head->srcNode) == KissBLE::ROUTER_NODEID)
			rv = 2;	// frame dst is router
		else
			rv = 4;	// other device on the same router (one broker)
	}
	else
		rv = 6;	// two or more brokers (routers)
	if(head->response || (head->noResponse && head->dstNode.connPort == 0))
		rv -= 1;
	return rv;
}

inline static NODE_STATE *getLocalNodeState(NODE_ID node)
{
	short idx = getLocalNode(node);
	return localNodes.getCellContent(idx);
}

static inline bool isRegistered(short devidx)
{
	if(devidx == thisRouterId)
		return true;
	NODE_STATE *ns = localNodes.getCellContent(devidx);
	return ((ns) ? (ns->registered) : false);
}

short getLocalNode(MAC_ADDR *mac)
{
	NODE_STATE *ns = NULL;
	short i;
	
	while((i = localNodes.nextBusyCell(&ns)) >= 0)
	{
		if(KissBLE::macEQ(mac, &ns->mac))
			return i;
	}
	return -1;
}

static short append2body(FRAME *frame, uint16_t len, const void *data = NULL)
{
	uint8_t *start;
	short nlen, blen, minLen = KissBLE::MIN_FRAME_LENGTH;

	if(frame->shead.frameLen == minLen)
		blen = 0;
	else
		blen = frame->payload.bodyLen;
	nlen = blen + len;
	if(nlen > KissBLE::MAX_BODY_SIZE)
		return -__LINE__;
	if(data && len > 0)
	{
		start = frame->payload.body + blen;
		memcpy(start, data, len);
	}
	frame->payload.bodyLen = nlen;
	// recalculate frame length
	frame->shead.frameLen = minLen + nlen;
	return 0;
}

void netDefaultParams()
{
	ADVERT_PARAMS advParams = {0};
	
	estimatedFrameDelay = 0;	// uSec
	delayMeasurments = 1;
	bzero(&advParams.timeSchema, sizeof(advParams.timeSchema));
	advParams.timeSchema.tickTime = 25;	// 125 ms
	advParams.timeSchema.maxTickCount = 7;
	advParams.timeSchema.sendTick = 0;		// slot 0 is transmit tick
	advParams.timeSchema.deviceTick = 1;
	advParams.timeSchema.deviceChannel = 1;
	advParams.timeSchema.sendChannel = 1;
	advParams.deviceWakeupInterval = 1000;	// in ms
	netParams.advParams = advParams;
	netParams.transmitDelay = TRANSMIT_DELAY_US;
	netParams.minTransmitDelay = MIN_TRANSMIT_DELAY_US;
}

void stopJob()
{
	stopProcessing();
	currentPhase = KissBLE::PROCESS_DISABLED;
	LOGGER(logStopJob);
}

inline static bool registerPossible()
{
	return (localNodes.countBusy() < (localNodes.CACHE_SIZE - 2));
}

CONNECTION_STATE *getConnectionState(NODE_ID node, short lnode = 0)
{
	CONNECTION_STATE *cs = NULL;
	short i, rnode = lnode;
	if(rnode == 0)
		rnode = netParams.idParams.node.nodeId;
	if(isRegistered(rnode))
	{
		if(lnode == netParams.idParams.node.nodeId)
			lnode = 0;
		while((i = connections.nextBusyCell(&cs)) >= 0)
		{
			if(cs->virtualOwner == lnode && CONN_EQ(cs->connNode, node))
				return cs;
		}
	}
	return NULL;
}

CONNECTION_STATE *getConnectionState(short destNode, short lnode = 0)
{
	NODE_ID dnode = composeNodeId(destNode);
	return getConnectionState(dnode, lnode);
}

static short oldestFrame4stream(CONNECTION_STATE *st)
{
	COMMAND_STATE *cs = NULL;
	short i;
	FHEAD *head;
	NODE_ID connId = st->connNode;
	uint8_t diff, next = st->nextSendSeq;
	
	while((i = cmdCache.nextBusyCell(&cs)) >= 0)
	{
		head = getHeader(cs);
		if(st->virtualOwner == cs->virtualOwner && 
			CONN_EQ(connId, head->dstNode) && head->frameSeq)
		{
			diff = next - head->frameSeq;
			if(diff > KissBLE::MAX_FRAMES_PER_STREAM)
				return i;
		}
	}
	return 0;
}

static short checkNextSeq(CONNECTION_STATE *st, FHEAD *head)
{
	uint8_t diff = head->frameSeq - st->lastReceivedSeq;
	if(diff > 1)
		return -KissBLE::FRAME_OUT_OF_SEQUENCE;
	if(diff == 1)
		st->lastReceivedSeq++;
	return 0;
}

static short setNextSeq(FHEAD *head)
{
	CONNECTION_STATE *st = getConnectionState(head->dstNode, head->srcNode.nodeId);
	if(head->dstNode.connPort == 0 || st == NULL)
	{	// UDP connection or first (register) frame for virtual node
		frameSequence++;
		if(frameSequence == 0)
			frameSequence = 1;
		head->frameSeq = frameSequence;
	}
	else
	{	// Stream connection. Check max limits of active frames
		short cmdidx = oldestFrame4stream(st);
		if(cmdidx > 0)
		{
			LOGGER(logNextSeq, cmdidx);
			return -KissBLE::TOO_MANY_FRAMES_PER_STREAM;
		}
		st->nextSendSeq++;
		if(st->nextSendSeq == 0)
			st->nextSendSeq = 1;
		head->frameSeq = st->nextSendSeq;
	}
	return 0;
}

static void initHeader(FRAME *frame, short frtype = 0)
{
	FHEAD *shead = &frame->shead;
	short emul = shead->emulated;
	IDENTITY_PARAMS *sp = &netParams.idParams;
	if(frtype == 0)
		frtype = shead->frameType;
	bzero(frame, KissBLE::MIN_FRAME_LENGTH);
	shead->emulated = emul;
	shead->frameType = frtype;
	shead->pan = sp->pan;
	shead->frameLen = KissBLE::MIN_FRAME_LENGTH;	// empty payload!
#if defined(KISS_EMULATOR)
	if(emul)
	{	// only devices
		if(IS_DEVICE)
			shead->dstNode = masterRouterNode;
		else
			shead->dstNode = sp->node;	// this router
		fillEmulatedHeader(shead);
	}
	else
#endif		
	{
		if(frtype == KissBLE::CTRL_REGISTER || frtype == KissBLE::CTRL_ACK)
		{
			if(IS_DEVICE || !IS_REGISTERED)
				shead->dstNode = masterRouterNode;
		}
		shead->srcRouter = thisRouterId;
		shead->srcNode = sp->node;
	}
}

// ISR context
static short sendIframe(IFRAME *iframe)
{
	short st, channel, len;
	int delay;
	COMMAND_STATE *cs;
	FRAME *frame;
	FHEAD *head;
	COMM_SCHEMA *cschema = NULL;
	CTRL_RADVERT_BODY *abody = NULL;
	CTRL_LANTERN_BODY *cbody = NULL;
	uint16_t lastCorrector = 0;
	uint32_t correctedShiftTime = 0;
	uint8_t *dest = NULL, encrypted[KissBLE::MAX_PACKET_SIZE];
			
	if(radio->getState() == Transceiver::RX)
		radio->stopReceive();
	cs = cmdCache.getCellContent(iframe->cmdIdx);
	head = getHeader(cs);
	channel = iframe->channel;
	iframe->startTime = ROUTER_TIME();
	LOGGER(logSendReceiveFrame, iframe, 'S');
	len = head->frameLen;
	frame = (FRAME *) memPool.getAddress(cs->poolHandle);	// encoded or not!
	dest = (uint8_t *) frame;
	lastFrameSendStop = ROUTER_TIME();
	if(cs->codedLen == 0)
	{	// Frame not encrypted
		if(frame->shead.frameType == KissBLE::CTRL_ADVERTISEMENT)
		{
			abody = (CTRL_RADVERT_BODY *) frame->payload.body;
			cschema = &abody->routerSchema;
		} else if(frame->shead.frameType == KissBLE::CTRL_LANTERN)
		{
			cbody = (CTRL_LANTERN_BODY *) frame->payload.body;
			cschema = &cbody->routerSchema;
		}
		if(cschema)
		{
			lastFrameSendStop = ROUTER_TIME();
			correctedShiftTime = (lastFrameSendStop - tickStartTime) + sendTimeCorrector;
			cschema->tickStartShift = correctedShiftTime / KissBLE::SHIFT_SCALE;
			if(abody)
				abody->routerTime = lastFrameSendStop + sendTimeCorrector;
			else if(cbody)
				cbody->routerTime = lastFrameSendStop + sendTimeCorrector;
		}
		//TODO.  Serialization!
		if(securityMethod && securityManager)
		{
			bool reset = cschema || cs->changed;
			dest = encrypted;
			len = securityManager->encodeFrame((uint8_t *)frame, dest, reset);
			if(len < 0)
				return len;
		}
	}
	st = radio->send(dest, len, channel);
	if(st == 0 && cschema)
	{	// time compensation for router time and tick start shift sended in ADVERT and LANTERN
		lastCorrector = (ROUTER_TIME() - lastFrameSendStop);
		if(timeCorrectorCount >= 8)
			timeCorrectorCount = 7;
		sendTimeCorrector = sendTimeCorrector * timeCorrectorCount + lastCorrector;
		timeCorrectorCount++;
		sendTimeCorrector = sendTimeCorrector / timeCorrectorCount;
	}
	// Estimate time spent to prepare frame to send
	if(delayMeasurments >= 8)
		delayMeasurments = 7;
	delay = estimatedFrameDelay * delayMeasurments + (lastFrameSendStop - lastFrameSendStart);
	delayMeasurments++;
	estimatedFrameDelay = delay / delayMeasurments;
	// TODO. copy back encoded frame to cache?
	if(cschema)
	{
		LOGGER(logSendTimes, head, correctedShiftTime, lastFrameSendStop,
					sendTimeCorrector, lastCorrector);
	}
	return st;	// st == 0 means OK
}

// ISR context
static short sendFrame(short cmdIdx, short channel)
{
	short rv;
	IFRAME eframe = {0};
	COMMAND_STATE *cs = cmdCache.getCellContent(cmdIdx);
	int32_t remain;
	short ignoreChannel = 0;
	
	lastFrameSendStart = ROUTER_TIME();	// send delay statistics
	remain = currentSlotEnd - lastFrameSendStart;
	if(remain < (radio->timeOnAir(KissBLE::MAX_PACKET_SIZE) + estimatedFrameDelay))
		return 0;	// Not enought time to send frame
	LOGGER(stopOrIgnore, KissLogger::DBG_BEFORE_SEND, cs, &ignoreChannel);
	if(ignoreChannel)
	{
		cs->ignored = 1;
		channel = ignoreChannel;
	}
	eframe.channel = channel;
	eframe.cmdIdx = cmdIdx;
	eframe.timeSchema = true;
	if(cs->virtualOwner == 0)
	{
		FRAME *frame = getFrame(cs);
		cs->head.srcRouter = frame->shead.srcRouter = thisRouterId;
	}
	rv = sendIframe(&eframe);
	return rv;	// 0 == OK
}

static short setBody(FRAME *frame, const PAYLOAD *body = NULL)
{
	short size = KissBLE::MIN_FRAME_LENGTH;
	if(body)
	{
		size += body->bodyLen;
		if(size > KissBLE::MAX_PACKET_SIZE)
			return -1;
		frame->payload = *body;
		if(frame->payload.bodyType == 0)
			frame->payload.bodyType = frame->shead.frameType;
	}
	else
		frame->payload.bodyLen = frame->payload.bodyType = 0;
	frame->shead.frameLen = size;
	return size;
}

short getOwner(FHEAD *head)
{
	short owner = 0;
	if(head->srcRouter) // from router to device
		owner = getVirtualNode(head->dstNode);	
	return owner;
}

static inline uint16_t encodeInterval()
{
	return ((currentIntervalNo & 0x7FFF) << 1) | 1;	// last 15 bits
}

static bool shrinkable(FHEAD *head)
{
	short ft = head->frameType;
	bool rv = (ft == KissBLE::CTRL_ACK || (USER_FRAME(ft) && head->response && 
			head->noResponse && head->dstNode.connPort));
	return rv;
}

short putIntoCache(FRAME *cframe, uint8_t flags = KissBLE::GENERATED_FRAME, 
					FHEAD *phead = NULL, short packetLen = 0)
{
	COMMAND_STATE *cs;
	short ft, own, flen, rv = -KissBLE::NO_FREE_SPACE_IN_CACHE;
	uint8_t generated = (flags & KissBLE::GENERATED_FRAME) ? 1 : 0;
	uint8_t omit = (flags & KissBLE::OMITTED_FRAME) ? 1 : 0;
	uint8_t eos = (flags & KissBLE::END_OF_SESSION) ? 1 : 0;
	uint8_t *destination, *source;
	FHEAD *head;
	FRAME *dframe;
	uint32_t pmap, smap;
	
	if(cframe == NULL)
		return -KissBLE::NO_FRAME_DATA;
	head = (phead) ? phead : &cframe->shead;
	if(phead == NULL)
		packetLen = head->frameLen;
	ft = head->frameType;
	cs = cmdCache.put();
	if(cs)
	{
		rv = cmdCache.getCellId(cs);
		bzero(cs, sizeof(COMMAND_STATE));
		cs->startInt = encodeInterval();
		cs->endOfSession = eos;
		cs->source = generated;
		cs->omitted = omit;
		cs->wait4process = (flags & KissBLE::WAIT4PROCESS) ? 1 : 0;
		cs->changed = generated;
		if(generated)
		{
			NODE_ID owner = netParams.idParams.node;
			if(ft == KissBLE::CTRL_ADVERTISEMENT ||	ft == KissBLE::CTRL_LANTERN ||
					ft == KissBLE::CTRL_ACK)
				head->noResponse = 1;
			if(shrinkable(head))
			{	// Full size because body can be extend just before send
				packetLen = sizeof(FRAME);
			}
			if(getVirtualNode(head->srcNode) > 0)
			{
				head->srcRouter = 0;
				owner = head->srcNode;
			}
			else
				head->srcRouter = thisRouterId;
			// set sequential number
			if(ft != KissBLE::CTRL_ADVERTISEMENT && head->frameSeq == 0 && 
					ft != KissBLE::CTRL_LANTERN)
			{
				flen = setNextSeq(head);	// check if frames limit not exceeded
				if(flen < 0)
				{
					FREE_COMMAND(rv, 1);
					return flen;
				}
			}
			own = getVirtualNode(owner);
		}
		else // from network
			own = getOwner(head);
		if(!head->noResponse && head->response == 0)
			cs->wait4response = 1;
		cs->virtualOwner = (own > 0) ? own : 0;
		cs->head = *head;
		if(!omit)
		{	// copy to pool
			cs->poolHandle = memPool.reserve(packetLen);
			if(cs->poolHandle == 0)
			{
				FREE_COMMAND(rv, 2);
				return -KissBLE::NO_FREE_SPACE_IN_CACHE;
			}
			memPool.getMaps(&pmap, &smap);
			dframe = (FRAME *) getFrame(cs);
			destination = (uint8_t *) dframe;
			source = (uint8_t *) cframe;
			memcpy(destination, source, packetLen);
			if(dframe->shead.emulated)
			{
				dframe->shead.emulated = 0;
				flags |= KissBLE::EMULATED;
			}
			if(phead)
				cs->codedLen = packetLen;
		}
		LOGGER(runRegisterFrame, cs, flags, rv, pmap, smap);
	}
	return rv;
}

static short buildResponseFrame(FRAME *frame)
{
	NODE_ID wrk;
	FHEAD *head = &frame->shead;

	head->response = 1;
	// swap source and destination nodes
	if(head->frameType == KissBLE::CTRL_OPEN_CONNECTION)
	{
		MAC_ADDR mac;
		CTRL_CONNECTION_BODY *sb = (CTRL_CONNECTION_BODY *) frame->payload.body;
		mac = sb->srcMac;
		sb->srcMac = sb->dstMac;
		sb->dstMac = mac;
	}
	wrk = head->dstNode;
	head->dstNode = head->srcNode;
	head->srcNode = wrk;
	return 0;
}

static short startListen(short channel, short line)
{
	short state = 0, oldChannel, rv = 0;
	bool started = false;
	state = radio->getState(&oldChannel);
	started = !(state == radio->RX);
	if(!started)
	{
		started = (oldChannel != channel);
		if(started)
		{
			radio->stopReceive();
			rv = radio->receive(channel);
		}
	}
	else
		rv = radio->receive(channel);
	LOGGER(startListen);
	return rv;
}

static short frameAtDestination(COMMAND_STATE *cs)
{
	FHEAD *head = getHeader(cs);
	if(EMPTY_NODE(head->dstNode))
		return -1;
	short rv = getVirtualNode(head->dstNode);
	if(cs->source || cs->virtualOwner != rv)
		rv = -1;	// generated || different owner
	return rv;
}

inline static bool frameAtSource(FHEAD *head)
{
	bool rv = (getVirtualNode(head->srcNode) > 0 && head->srcRouter == 0);
	if(!rv)
		rv = THIS_NODE(head->srcNode);
	return rv;
}

static void processAfterSendFrame(IFRAME lastSended, short nextIdx, short stat)
{
	short shrinkSize = 0, ft, lastIdx = lastSended.cmdIdx;
	bool stop = false;
	COMMAND_STATE *cs = cmdCache.getCellContent(lastIdx);
	FHEAD *head = NULL;
	
	// Continue processing after send frame
	LOGGER(stopOrIgnore, KissLogger::DBG_AFTER_SEND, cs, NULL);
	switch2routerTime(&lastSended);
	if(lastIdx >= 0)
	{	// unmark sended frame.
		cs->frameSendInt = encodeInterval();	// last 15 bits
		head = getHeader(cs);
		ft = head->frameType;
		if(ft == KissBLE::CTRL_LANTERN || ft == KissBLE::CTRL_ADVERTISEMENT)
		{
			NODE_STATE *ns = getLocalNodeState(netParams.idParams.node);
			if(ns)
				ns->nodeActivity = encodeInterval();
			if(ft == KissBLE::CTRL_LANTERN && !IS_DEVICE)
			{
				head->pan = 0;
			}	// else device lantern frame permanently in cache
		}
		else 
		{
			LOGGER(incrementFrames);
		}
		if(cs->codedLen == 0 && shrinkable(head))
			shrinkSize = memPool.shrink(cs->poolHandle, head->frameLen);	// Shrink after send
		LOGGER(logFrameTime, &lastSended, shrinkSize);
#if defined(KISS_EMULATOR)
		if(!IS_DEVICE)
			emulateVirtualResponse(lastIdx);
#endif
		// Remove immediate after send?
		if(lastIdx > KissBLE::RESERVED_CACHE_CELL_IDX)
		{
			bool remove = (ft == KissBLE::CTRL_LANTERN || ft == KissBLE::CTRL_ACK);
			if(!remove)
				remove = (ft == KissBLE::CTRL_REGISTER && head->response);
			if(remove)
			{
				FREE_COMMAND(lastIdx, 3);
			}
		}
		LOGGER(stopTestSessions, 'S', cs, &stop);
		if(stop)
			return;
	}
	// next frame sended?
	if(nextIdx < 0)
	{
		LOGGER(markSendStop, &lastSended);
#if defined(KISS_EMULATOR)
		{
			if(!IS_DEVICE)
				emulate(currentIntervalNo);
		}
#endif
	}
	else
	{
		LOGGER(logFrameError, stat);
	}
	(void) shrinkSize;	// warning
}

// ISR context
static short sendNextFrame(short lastStatus)
{
	short nextIdx = -1, stat = 0;
	IFRAME iframe = {0};
	if(lastStatus == 0)
	{	// frame emited. Uncheck
		iframe.cmdIdx = getSetBit(frames2sendMap);
		iframe.len = radio->getPacketInfo(&iframe.startTime, &iframe.endTime, false);
		if(iframe.cmdIdx >= 0)
			clearLeftBit(&frames2sendMap, iframe.cmdIdx); // unmark sended frame.
		LOGGER(incrementSent);
	}
	else
		iframe.cmdIdx = -1;
	if(currentSlotCompletedTime == 0)
	{	// next frame to send?
		nextIdx = getSetBit(frames2sendMap);
		if(nextIdx >= 0)
			stat = sendFrame(nextIdx, CHANNEL(currentSlotChannel));
		else
		{
			currentSlotCompletedTime = (iframe.endTime) ? iframe.endTime : ROUTER_TIME();
			if(IS_DEVICE || !IS_REGISTERED)
				radio->stopReceive();
			else
				startListen(CHANNEL(currentSlotChannel), __LINE__);
		}
	}
	kissQueue->call(processAfterSendFrame, iframe, nextIdx, stat);	// leave ISR context
	return stat;
}

static short bit2seq(uint32_t mask, short node)
{
	uint32_t wrk = 0;
	
	setLeftBit(&wrk, node);
	if(wrk & mask)
	{
		wrk -= 1;
		wrk = ~wrk;
		wrk &= mask;
		return countSetBits(wrk);
	}
	return -1;
}

uint64_t nodeTransmitGapParams(uint64_t *stop, short nodeNo = -1)
{
	short devCount, nodeSeq;
	uint32_t rmask;
	uint64_t slotStart, slotTime, gapTime, gap1time;
	CTRL_RADVERT_BODY *ab;
	
	*stop = 0;
	if(thisRouterId)
	{
//		COMMAND_STATE *cs = cmdCache.getCellContent(KissBLE::RESERVED_CACHE_CELL_IDX);
		FRAME *frame = getFrame(KissBLE::RESERVED_CACHE_CELL_IDX);
		ab = (CTRL_RADVERT_BODY *) frame->payload.body;
	}
	else
	{
		if(currentSlot != activeSchema.sendTick && !IS_REGISTERED)
			return 0;
		ab = &routerAdvertisement;
	}
	if(currentSlot == activeSchema.sendTick)
	{
		slotStart = currentSlotCompletedTime;
		slotTime = (currentSlotEnd - slotStart);
		rmask = ab->extraTimeNodes << 16;
		if(ab->registerActive)
			setLeftBit(&rmask, KissBLE::REGISTER_GAP_IDX);
	}
	else if(currentSlot == activeSchema.deviceTick)
	{	// deviceTick
		slotTime = TICK_TIME_US;
		slotStart = tickStartTime + currentSlot * slotTime; // in uS
		rmask = ab->nodes2send << 16;
		if(ab->routerGap)
			insertBit(&rmask, ab->routerGap);
	} 
	else
	{
		LOGGER(logMessage, LOG_LEVEL_MINIMUM, 
			 "*** nodeTransmitGap not in proper timeslot (%d)\n", currentSlot);
		return 0;
	}
	if(nodeNo < 0)
	{
		if(IS_REGISTERED)
			nodeNo = getLocalNode(netParams.idParams.node);
		else
			nodeNo = KissBLE::REGISTER_GAP_IDX;
	}
	devCount = countSetBits(rmask);
	nodeSeq = bit2seq(rmask, nodeNo);
	if(devCount <= 0 || nodeSeq < 1)
	{
		LOGGER(logMessage, LOG_LEVEL_FULL, 
			"Gap in slot %d for node %d disabled (%d/%d) nodeMask:%8.8lX\n", 
			currentSlot, nodeNo, nodeSeq, devCount, rmask);
		return 0;
	}
	gapTime = slotTime / (devCount * 1000);	// in ms
	gapTime *= 1000;
	// first gap time is a little longer
	gap1time = gapTime + (slotTime - devCount * gapTime);	// in uS
	// first transmition delayed 
	slotStart += netParams.transmitDelay;
	gap1time -= netParams.transmitDelay;
	if(nodeSeq == 1)
		*stop = slotStart + gap1time;
	else
	{
		slotStart += gap1time + (nodeSeq - 2) * gapTime;
		*stop = slotStart + gapTime;
	}
	LOGGER(logGapParams, nodeNo, nodeSeq, devCount, rmask, slotTime, gap1time, 
			gapTime, slotStart, *stop);
	return slotStart;
}

static uint16_t computeNodes2send()
{
	uint16_t rv = localNodes.getBitmap() >> 16;
	rv &= 0x03FFF;	// bits ZERO and ONE are not registered devices
	return rv;	// every registered node can transmit
}

static uint16_t computeExtraNode2send()
{
	uint16_t rv = localNodes.getBitmap() >> 16;
	rv &= 0x03FFF;
	return rv;	// every registered node can transmit
}

static short fillAdvBody(FRAME *frame)
{
	uint16_t nodes = computeNodes2send();
	CTRL_RADVERT_BODY *abody = (CTRL_RADVERT_BODY *) &frame->payload.body;
	
	abody->routerGap = 1;
	if(randomData)
	{
		uint8_t	r;
		short st = randomData(&r, sizeof(r));
		if(st > 0)
			abody->routerGap = (r % 7) + 1;
	}
	if(registerPossible())
		abody->canRegister = 1;
	abody->registerActive = 1;
	abody->routerSchema = activeSchema;
	abody->nodes2send = nodes;
	abody->extraTimeNodes = computeExtraNode2send();
	abody->reservePriority = netParams.idParams.reservePriority;
	return sizeof(CTRL_RADVERT_BODY);
}

static short fillRegisterBody(FRAME *frame)
{
#if defined(KISS_EMULATOR)
	if(frame->shead.emulated)
		fillEmulatedBody(frame);
	else
#endif		
	{
		CTRL_REGISTER_BODY *cbody = (CTRL_REGISTER_BODY *) frame->payload.body;
		if(netParams.idParams.mac.len == 0)
		{
			LOGGER(logMessage, LOG_LEVEL_MINIMUM, "*** MAC address unknown.\n");
			KissBLE::doJob(-1);
			return 0;
		}
		cbody->mac = netParams.idParams.mac;
	}
	return sizeof(CTRL_REGISTER_BODY);
}

static short fillOpenConnectionBody(FRAME *frame, void *extraPar)
{
	MAC_ADDR *dmac = (MAC_ADDR *) extraPar;
	CTRL_CONNECTION_BODY *session = (CTRL_CONNECTION_BODY *) frame->payload.body;
#if defined(KISS_EMULATOR)
	if(frame->shead.emulated)
		fillEmulatedBody(frame);
	else
#endif		
	{
		const char *msg = NULL;
		if(extraPar == NULL)
		{
			msg = "*** MAC address not specified.\n";
		}
		session->dstMac = *dmac;
		session->srcMac = netParams.idParams.mac;
		if(thisRouterId)
		{
			short rnode = getLocalNode(dmac);
			if(rnode > 0 && isRegistered(rnode))
				frame->shead.dstNode = composeNodeId(rnode);
			else
			{	// TODO. Check if other (sub)routers active
				msg = "*** Node with MAC address not registered\n";
			}
		}
		if(msg)
		{
			LOGGER(logMessage, LOG_LEVEL_MINIMUM, msg);
			return -1;
		}
	}
	return sizeof(CTRL_CONNECTION_BODY);
}

static short fillLanternBody(FRAME *frame)
{
	CTRL_LANTERN_BODY *cbody = (CTRL_LANTERN_BODY*) frame->payload.body;
	cbody->routerSchema = activeSchema;
	cbody->reservePriority = netParams.idParams.reservePriority;
	return sizeof(CTRL_LANTERN_BODY);
}

static NODE_ID router4node(NODE_ID node = netParams.idParams.node)
{
	short vn;
	NODE_ID rv = KissBLE::nullNodeId;
	if(EMPTY_NODE(node))
		return KissBLE::nullNodeId;
	vn = getLocalNode(node);
	if(vn != KissBLE::ROUTER_NODEID)
	{
		if(IS_DEVICE)
			rv = masterRouterNode;
		else
			rv = netParams.idParams.node;	// this is router
	}
	return rv;
}

static short fillACKbody(FRAME *frame, void *extraPar)
{
	HEAD_LIST *cbody = (HEAD_LIST *) frame->payload.body;
	short rv = sizeof(cbody->headListLen);
	if(extraPar)
	{
		FHEAD *head = (FHEAD *) extraPar, *newHead = &frame->shead;
		short vn = getVirtualNode(head->dstNode);
		cbody->headList[0] = *head;
		cbody->headListLen = 1;
		newHead->srcRouter = (vn > 0) ? 0 : thisRouterId;
		newHead->srcNode = head->dstNode;
		newHead->dstNode = (head->srcRouter) ? router4node(head->dstNode) : head->srcNode;
		newHead->dstNode.connPort = newHead->srcNode.connPort = 0;
		rv += sizeof(cbody->headList[0]);
	}
	return rv;
}

static short fillUserBody(FRAME *frame, void *extraPar)
{
	short rv = 0;
	PAYLOAD *pay = &frame->payload;
	
	if(extraPar)
	{
		rv = strlen((char *) extraPar);
		memcpy(pay->body, extraPar, rv);
	}
#if defined(KISS_EMULATOR)	
	if(frame->shead.emulated)
	{
		rv = fillEmulatedBody(frame);
	}
#endif
	return rv;
}

static short composeFrame(FRAME *frame, void *extraPar = NULL)
{
	short bodyLen = 0;
	
	initHeader(frame);
	frame->payload.bodyType = frame->shead.frameType;
	switch(frame->shead.frameType)
	{
		case KissBLE::CTRL_ADVERTISEMENT:
			bodyLen = fillAdvBody(frame);
			break;
		case KissBLE::CTRL_REGISTER:
			bodyLen = fillRegisterBody(frame);
			break;
		case KissBLE::CTRL_OPEN_CONNECTION:
			bodyLen = fillOpenConnectionBody(frame, extraPar);
			break;
		case KissBLE::CTRL_LANTERN:
			bodyLen = fillLanternBody(frame);
			break;
		case KissBLE::CTRL_ACK:
			bodyLen = fillACKbody(frame, extraPar);
			break;
		case KissBLE::USER_STREAM_BYTE:
			bodyLen = fillUserBody(frame, extraPar);
			break;
	}
	if(bodyLen < 0)
		return bodyLen;
	if(append2body(frame, bodyLen) < 0)
		return -1;
	return 0;
}

short registerFrame(FRAME *frame, void *extraPar = NULL)
{
	short rv = composeFrame(frame, extraPar);
	if(rv >= 0)
		rv = putIntoCache(frame);
	return rv;
}

short registerFrame(short ctrlType, void *extraPar = NULL)
{
	FRAME frame = {0};
	frame.shead.frameType = ctrlType;
	return registerFrame(&frame, extraPar);
}

// ISR context
static void startTransmit()
{
	currentSlotCompletedTime = 0;
	sendNextFrame(-1);
	LOGGER(logTransmissionStart, frames2sendMap);
}

static short startOrPlanTransmit(uint64_t sendStart, uint64_t sendStop)
{
	short extraMess = 0;
	int32_t rtime = 0;
	uint64_t ct = ROUTER_TIME();
	if(ct >= sendStart && ct < sendStop)
	{	// we are in proper transmit time slot/gap
		rtime = ct - (sendStop - radio->timeOnAir(KissBLE::MAX_PACKET_SIZE));
		extraMess = (rtime < 0) ? 1 : 2;
	}
	else if(ct < sendStart)
	{
		rtime = sendStart - ct;	// time to wait
		if(rtime < netParams.minTransmitDelay)
			rtime = netParams.minTransmitDelay;
		extraMess = 3;
	} 
	else
		extraMess = 4;
	LOGGER(logPlanTransmit, extraMess, rtime, sendStart, sendStop);
	if(rtime < 0)
		startTransmit();
	else if(rtime > 0)
		transmitTimeControl.attach_us(startTransmit, rtime);
	return extraMess;
}

static void randomShiftStart(uint64_t *start, uint64_t stop)
{
	int32_t fcount;
	uint16_t fl;
	short rv, st;
	
	if(*start <= 0)
		return;
	fl = radio->timeOnAir();
	fcount = ((stop - *start) / fl) - 2;
	if(fcount < 2)
		return;
	if(randomData)
	{
		st = randomData((uint8_t *) &rv, sizeof(rv));
		if(st > 0)
		{
			rv = rv % fcount;
			fcount = rv * fl;
			*start += fcount;
			LOGGER(logMessage, LOG_LEVEL_FULL, 
				"Register frame random shift gap %d.%3.3d ms\n",
				fcount / 1000, fcount % 1000);
		}
	}
}

// Not ISR context
static short prepareSend4device(uint64_t *sendStart, uint64_t *sendStop)
{	// first cell permanently occupied by lantern frame!
	short i = -1;
	uint16_t ct;
	COMMAND_STATE *cs;
	
	if(!ADVERT_ACTIVE)
		return -1;
	// Node can transmit in this gap?
	*sendStart = nodeTransmitGapParams(sendStop);
	if(!IS_REGISTERED)
	{
		FRAME *frame;
		if(*sendStart <= 0 || currentSlot != activeSchema.sendTick)
		{
			*sendStart = 0;
			return -1;
		}
		// Register frame temporary in cache
		cs = cmdCache.getCellContent(KissBLE::RESERVED_CACHE_CELL_IDX + 1);
		frame = getFrame(cs, true);	// not yet encrypted?
		if(frame)
		{
			frame->shead.dstNode = masterRouterNode;	// first time
		}
		if(randomData)
			randomShiftStart(sendStart, *sendStop);
		return KissBLE::RESERVED_CACHE_CELL_IDX;	// omit lantern frame
	}
	// Can transmit in this timeslot interval?
	if(*sendStart <= 0)
		*sendStart = gap4virtualNode(sendStop);
	// Send lantern frame?
	cs = cmdCache.getCellContent(KissBLE::RESERVED_CACHE_CELL_IDX);
	if(cs->frameSendInt)
	{
		ct = currentIntervalNo & 0x7FFF;
		ct = (ct - KissBLE::decodeInterval(cs->frameSendInt)) * INTERVAL_TIME;
		if(ct <= netParams.advParams.deviceWakeupInterval)
			i = KissBLE::RESERVED_CACHE_CELL_IDX;	// skip lantern (first) frame
		else
			cs->frameSendInt = 0;
	}
	return i;
}

static short prepareSend4router(uint64_t *sendStart, uint64_t *sendStop)
{
	COMMAND_STATE *cs = cmdCache.getCellContent(KissBLE::RESERVED_CACHE_CELL_IDX);
	FRAME *frame = getFrame(cs);
	cs->frameSendInt = 0;
	fillAdvBody(frame);
	// sending start time 
	*sendStart = tickStartTime + activeSchema.sendTick * TICK_TIME_US;
	*sendStop = *sendStart + TICK_TIME_US;
	*sendStart += netParams.transmitDelay;
	return -1;
}

static short findFrame4response(FHEAD *head, short owner)
{
	short i;
	HEAD_LIST *cb;
	COMMAND_STATE *cs = NULL;
	FRAME *frame;
	
	while((i = cmdCache.nextBusyCell(&cs)) >= 0)
	{
		if(cs->codedLen)
			continue;
		frame = getFrame(cs);
		if(frame && USER_FRAME(frame->shead.frameType) && cs->virtualOwner == owner &&
			frame->shead.response && NODE_EQ(frame->shead.dstNode, head->srcNode) &&
			frame->payload.bodyType == KissBLE::CTRL_HEAD_LIST && cs->frameSendInt == 0)
		{
			cb = (HEAD_LIST *) frame->payload.body;
			if(cb->headListLen < KissBLE::ACK_LIST_SIZE)
				return i;
		}
	}
	return -1;
}

short pushResponse(short cmdidx, PAYLOAD *body = NULL)
{
	short respidx;
	FRAME response = {0};
	COMMAND_STATE *cs = cmdCache.getCellContent(cmdidx);
	FHEAD *head = getHeader(cs);
	PAYLOAD ubody;

	if(!cs->wait4response)
	{
		if(USER_FRAME(head->frameType) && head->dstNode.connPort && 
			head->noResponse)
		{	// stream port and no response. Clear body. Send empty (multi)response
			HEAD_LIST *list;
			response.shead = *head;
			respidx = findFrame4response(head, cs->virtualOwner);
			if(respidx >= 0)
			{
				cs = cmdCache.getCellContent(respidx);
				FRAME *frame2 = getFrame(cs);
				list = (HEAD_LIST *) frame2->payload.body;
				buildResponseFrame(&response);
				append2body(frame2, sizeof(FHEAD), &response.shead);
				list->headListLen++;
				frame2->shead.userMultiResponse = 1;
				LOGGER(logPushResponse, head, &response.shead, cs);
				FREE_COMMAND(cmdidx, 4);
				return respidx;
			}
			setBody(&response);
			ubody.bodyType = KissBLE::CTRL_HEAD_LIST;
			list = (HEAD_LIST *) ubody.body;
			ubody.bodyLen = sizeof(list->headListLen);
			list->headListLen = 0;
			body = &ubody;
		}
		else
		{
			LOGGER(logMessage, LOG_LEVEL_MINIMUM, 
				 "*** Response not allowed (noResponse flag) for frame[%d]\n", 
				 cmdidx);
			return -KissBLE::RESPONSE_NOT_ALLOWED;
		}
	}
	else
	{
		FRAME *frame = getFrame(cs);
		if(frame == NULL)
			return -KissBLE::FRAME_SECURITY_BROKEN;
		memcpy(&response, frame, KissBLE::getLength(frame));
	}
	if(body)
	{
		if(setBody(&response, body) < 0)
			return -1;
	}
	buildResponseFrame(&response);
	respidx = putIntoCache(&response);
	if(respidx >= 0)
	{
		FREE_COMMAND(cmdidx);
	}
	return respidx;
}

static short checkTimeouts(short cmd)
{
	short h = 0, rv = -1, mx;
	COMMAND_STATE *cs = cmdCache.getCellContent(cmd);
	short intervals = (currentIntervalNo & 0x7FFF) - KissBLE::decodeInterval(cs->frameSendInt);
	FHEAD *head = getHeader(cs);
	bool vframe = (getVirtualNode(head->srcNode) > 0 || 
					getVirtualNode(head->dstNode) > 0);
	// emulated nodes process on router is delayed
	mx = (vframe && !IS_DEVICE) ? 1 : 0;
	if(cs->ackReceived)
	{	// wait for response?
		h = frameHoops(head);
		mx += h;
		if(intervals > mx)
		{
			uint16_t sdiff = (currentIntervalNo & 0x7FFF) - KissBLE::decodeInterval(cs->startInt);
			if(sdiff > 2 * mx)
			{
				if(cs->wait4response && !head->response)
				{	// Send error frame to source node
					head->netStatus = KissBLE::PATH_TO_NODE_BROKEN;
					rv = pushResponse(cmd);
					if(rv >= 0)
						cs = cmdCache.getCellContent(rv);	// for logging below
				}
				else
				{
					FREE_COMMAND(cmd, 5); // remove response
					return -1;
				}
			}
			else
			{
				cs->ackReceived = 0;	// resend
				rv = cmd;
			}
		}
	}
	else if(frameAtDestination(cs) && frameHoops(head) <= 1)
	{
		FREE_COMMAND(cmd, 6);
		return -1;
	}
	else
	{
		mx += 1;
		if(intervals > mx)
			rv = cmd;
	}
	LOGGER(logTimeout, cs, intervals, mx, h, rv);
	return rv;
}

static short findActiveACK(short owner)
{
	short i;
	COMMAND_STATE *cs = NULL;
	HEAD_LIST *cb;
	FRAME *frame;
	while((i = cmdCache.nextBusyCell(&cs)) >= 0)
	{
		if(cs->codedLen)
			continue;	// skip encrypted (and sended)
		frame = getFrame(cs);
		if(frame && frame->shead.frameType == KissBLE::CTRL_ACK && 
			cs->virtualOwner == owner && cs->frameSendInt == 0)
		{
			cb = (HEAD_LIST *) frame->payload.body;
			if(cb->headListLen < KissBLE::ACK_LIST_SIZE)
				return i;
		}
	}
	return -1;
}

static short append2ack(FHEAD *head)
{
	short vn, ack = -1;
	
	if(frameHoops(head) <= 1)
		return false;	// short path frames or router never generate ACK frame
	vn = getVirtualNode(head->dstNode);
	if(vn >= 0)
	{
		ack = findActiveACK(vn);
		if(ack < 0)
			ack = registerFrame(KissBLE::CTRL_ACK, head);
		else
		{
			COMMAND_STATE *cs = cmdCache.getCellContent(ack);
			FRAME *frame = getFrame(cs);
			if(frame)
			{
				HEAD_LIST *ackb;
				append2body(frame, sizeof(FHEAD), head);
				ackb = (HEAD_LIST *) frame->payload.body;
				ackb->headListLen++;
			}
		}
		LOGGER(logAppendHead, ack, head);
	}
	return ack;
}

// Not ISR context
static void prepare4send()
{
	uint64_t sendStart, sendStop;
	int16_t i, last, first;
	uint16_t localNodeMask;
	COMMAND_STATE *cs = NULL, *csi;
	FRAME *frame;
	FHEAD *head;
	CTRL_RADVERT_BODY *abody = NULL;
	
	frames2sendMap = 0;
	transmitTimeControl.detach();
	if(currentPhase == KissBLE::PROCESS_DISABLED)
		return;
	LOGGER(logEntryPrepare);
	LOGGER(stopOrIgnore, KissLogger::DBG_PREPARE, NULL, NULL);
	sendStart = sendStop = 0;
	if(!thisRouterId)
		i = prepareSend4device(&sendStart, &sendStop);
	else
		i = prepareSend4router(&sendStart, &sendStop);
	if(sendStart <= 0)
	{
		LOGGER(logMessage, LOG_LEVEL_FULL, "Sending not active in current timeslot\n");
		return;
	}
	csi = cs = (i < 0) ? NULL : cmdCache.getCellContent(i);
	localNodeMask = 0;
	last = -1;
	while((i = cmdCache.nextBusyCell(&cs)) >= 0)
	{
		if(cs->omitted || cs->wait4process)
			continue;
		head = getHeader(cs);
		cs->ignored = cs->endOfSession = 0;
		if(cs->frameSendInt)
			i = checkTimeouts(i); // sended
		else if(frameAtDestination(cs) >= 0)
		{	// ACK must be send?
			i = -1;
			if(cs->ackSendInt == 0)
			{
				i = append2ack(head);
				if(i >= 0)
				{
					cs->ackSendInt = encodeInterval();
					csi = cmdCache.getCellContent(i);
				}
			}
			else if(head->response && !cs->wait4handler)
			{	// remove after send ACK
				uint16_t delay = (currentIntervalNo & 0x7FFF) - KissBLE::decodeInterval(cs->ackSendInt);
				if(delay >= KissBLE::WAIT4ACK_TIME)
					FREE_COMMAND(i, 7);
			}
		}	// else will be (re)send
		if(i >= 0 && !csi->wait4handler)
		{
			setLeftBit(&frames2sendMap, i);
			if(thisRouterId)
			{	// monitor EOS for frame flag (last in stream) and bitmap devices waiting for transmition
				if(i > last)
					last = i;
				getLocalNode(head->dstNode, &localNodeMask);
			}
		}
	}
	first = getSetBit(frames2sendMap);
	if(first >= 0)
	{	// Something to send
		// First, reserved cell?
		if(first == KissBLE::RESERVED_CACHE_CELL_IDX)
		{	// reserved ADVERTISEMENT or LANTERN frame
			cs = cmdCache.getCellContent(KissBLE::RESERVED_CACHE_CELL_IDX);
			frame = getFrame(cs);
			frame->shead.frameSeq++;
			if(thisRouterId)
			{
				csi = cmdCache.getCellContent(last);
				csi->endOfSession = 1;
				abody = (CTRL_RADVERT_BODY *) frame->payload.body;
				abody->nodes2wait = localNodeMask;
				head = getHeader(csi);
				abody->eosId = composeEOS(head);
			}
		}
	}
	LOGGER(logExitPrepare, frames2sendMap);
	if(first >= 0)
	{	// buffer first packet into transceiver cache
		radio->getPacket(NULL, 0);	// clear transmit kissQueue
		startOrPlanTransmit(sendStart, sendStop);
	}
}

static bool shiftTickStart()
{
	uint64_t dt, ct, nstart;
	short cslot, ns, tphase;
	uint32_t usTick, interval = INTERVAL_TIME * 1000;
	uint32_t cinterval;
	bool rv; //asDev = IS_DEVICE || !IS_REGISTERED, rv;
	
	ct = ROUTER_TIME();
	cinterval = (ct - netStartTime) / interval;
	dt = (ct - schemaSyncTime) / interval;
	tickStartTime = schemaSyncTime + dt * interval;
	if(thisRouterId)
		schemaSyncTime = tickStartTime; // self synchronization
	usTick = TICK_TIME_US;
	cslot = (ct - tickStartTime) / usTick;
	nstart = tickStartTime + cslot * usTick;
	rv = (cinterval != currentIntervalNo || cslot != currentSlot);
	if(rv || nstart != currentSlotStart)
	{
		currentIntervalNo = cinterval;
		currentSlot = cslot;
		currentSlotStart = nstart;
		currentSlotCompletedTime = 0;
		currentSlotChannel = slotModeTab[cslot].slotChannel;
		tphase = currentPhase = slotModeTab[cslot].phase;
		ns = 0;
		cslot = currentSlot + 1;
		while(cslot != currentSlot)
		{
			cslot %= activeSchema.maxTickCount + 1;
			if(slotModeTab[cslot].phase != tphase || 
					slotModeTab[cslot].slotChannel != currentSlotChannel)
				break;
			ns++;
			cslot++;
		}
		currentSlotEnd = tickStartTime + (currentSlot + 1 + ns) * usTick;
	}
	return rv;
}

static void clearGlobals()
{
	transmitTimeControl.detach();
	newPhaseTimeControl.detach();
	radio->stopReceive(true);
	radio->getPacket(NULL, 0);
	currentSlot = -1;
	sendTimeCorrector = timeCorrectorCount = 0;
	currentIntervalNo = thisRouterId = currentSlotChannel = cmdPending4handler = 0;
	syncLoop = 0;
	currentSlotEnd = currentSlotCompletedTime = 0;
	netParams.idParams.node = masterRouterNode = KissBLE::nullNodeId;
	netStartTime = schemaSyncTime = tickStartTime = 0;
	frameSequence = 0;
	localNodes.clear();
	cmdCache.clear();
	bzero(&activeSchema, sizeof(activeSchema));
	LOGGER(clear);
	setScenarioInterval(-1, -1);
	connections.clear();
	memPool.reset();
}

static void fillSlotsTable()
{
	short i, mx;
	bool spec = thisRouterId;
	
	mx = sizeof(slotModeTab) / sizeof(slotModeTab[0]);
	DISABLE_ISR;
	for(i = 0; i < mx; i++)
	{
		if(spec)
		{
			slotModeTab[i].phase = KissBLE::LISTEN_PHASE;
			slotModeTab[i].slotChannel = KissBLE::SYNC_ROUTER_CHANNEL_IDX;
		}
		else
		{
			slotModeTab[i].phase = KissBLE::IDLE_PHASE;
			slotModeTab[i].slotChannel = 0;
		}
	}
	slotModeTab[activeSchema.sendTick].slotChannel = activeSchema.sendChannel;
	slotModeTab[activeSchema.sendTick].phase = (spec) ? KissBLE::SEND_PHASE : KissBLE::LISTEN_PHASE;
	slotModeTab[activeSchema.deviceTick].slotChannel = activeSchema.deviceChannel;
	slotModeTab[activeSchema.deviceTick].phase = (spec) ? KissBLE::LISTEN_PHASE : KissBLE::SEND_PHASE;
}

static void startAsSpecialNode()
{
	NODE_STATE *ns;
	
	if(tickStartTime == 0)
		netStartTime = tickStartTime = schemaSyncTime = ROUTER_TIME();
	netParams.idParams.node.role = netParams.idParams.reserveRole;
	thisRouterId = netParams.idParams.node.nodeId = KissBLE::ROUTER_NODEID;
	netParams.idParams.node.routerId = KissBLE::SYNC_ROUTER_NET_SLOT;
	localNodes.put(NULL, 0);	// ignored
	ns = localNodes.put(NULL, KissBLE::ROUTER_NODEID);
	if(ns)
		ns->mac = netParams.idParams.mac;
	setScenarioInterval(0, netParams.idParams.node.nodeId);
	activeSchema = netParams.advParams.timeSchema;
	fillSlotsTable();
	cmdCache.clear();
	registerFrame(KissBLE::CTRL_ADVERTISEMENT);
	LOGGER(logStart);
	setCurrentPhase(NULL);
}

// Only for special (sync) router !
static void prepareLantern4send()
{
	uint64_t sendStart, sendStop;
	COMMAND_STATE *cs = cmdCache.getCellContent(KissBLE::RESERVED_CACHE_CELL_IDX);
	FRAME *frame = getFrame(cs);
	CTRL_RADVERT_BODY *adv = (CTRL_RADVERT_BODY *) frame->payload.body;
	
	sendStart = nodeTransmitGapParams(&sendStop, adv->routerGap);
	LOGGER(logPrepareLantern, sendStart, sendStop, adv);
	if(sendStart > 0)
	{
		uint32_t diff, rshift, toa;
		short st, i = registerFrame(KissBLE::CTRL_LANTERN);
		if(i < 0)
			return;
		frames2sendMap = 0;
		setLeftBit(&frames2sendMap, i);
		radio->getPacket(NULL, 0);	// clear transmit kissQueue in radio
		// Random start in timeslot for lantern.
		toa = radio->timeOnAir();
		diff = (sendStop - sendStart) / toa;
		diff -= 2;	// last 2 frames times discarded
		if(diff > 0)
		{
			if(randomData)
			{
				st = randomData((uint8_t *) &rshift, sizeof(rshift));
				if(st > 0)
					sendStart += (rshift % diff) * toa;
			}
		}
		startOrPlanTransmit(sendStart, sendStop);
	}
}

// timer ISR routine
static void setCurrentPhaseISR()
{
	setCurrentPhase(NULL);
}

// ISR context if advFrame is NULL
static void setCurrentPhase(FRAME *advFrame)
{
	uint64_t stime, startProc;
	uint32_t oldInterval;
	short channel;
	uint16_t toa;
	static uint16_t minStep;
	bool slotChanged;
	
	if(currentPhase == KissBLE::PROCESS_DISABLED || !SCHEMA_ACTIVE)
		return;
	startProc = ROUTER_TIME();
	if(currentPhase == KissBLE::LISTEN_PHASE && tickStartTime != schemaSyncTime)
	{
		if(!thisRouterId)
		{
			LOGGER(logMessage, LOG_LEVEL_FRAMES, "Missing advertisement\n");
			routerInactiveCount++;
			if(routerInactiveCount >= ROUTER_INACTIVE_MAX)
			{
				kissQueue->call(syncPhase);
				return;
			}
		}
	}
	// Dynamic logging
	oldInterval = currentIntervalNo;
	slotChanged = shiftTickStart();
	if(oldInterval != currentIntervalNo)
	{
		LOGGER(checkStopRun);
	}
	LOGGER(currentPhaseStat, slotChanged, minStep);
	stime = currentSlotEnd - ROUTER_TIME();
	if(stime >= radio->timeOnAir(KissBLE::MAX_PACKET_SIZE))
	{	// Its enought time to interact with radio
		channel = CHANNEL(currentSlotChannel);
		if(currentPhase == KissBLE::LISTEN_PHASE)
		{
			if(slotChanged)
			{
				bool listen = true;
				if(thisRouterId)
				{
					if(currentSlot == activeSchema.deviceTick)
					{	// set time of  transmit of router lantern frame
						COMMAND_STATE *cs = cmdCache.getCellContent(KissBLE::RESERVED_CACHE_CELL_IDX);
						FRAME *frame = getFrame(cs);
						CTRL_RADVERT_BODY *adv = (CTRL_RADVERT_BODY *) frame->payload.body;
						if(adv->routerGap)
							kissQueue->call(prepareLantern4send);
					}
				} 
				else if(currentSlot == activeSchema.sendTick && advFrame)
				{	// if device and Advertisement with EOS then not start listen!
					CTRL_RADVERT_BODY *adv = (CTRL_RADVERT_BODY *) advFrame->payload.body;
					if(checkEOS(&advFrame->shead, adv->eosId))
						listen = false;
				}
				if(listen)
					startListen(channel, __LINE__);
			}
		}
		else if(currentPhase == KissBLE::SEND_PHASE)
		{
			if(slotChanged)
				kissQueue->call(prepare4send);
		}
		else
			radio->stopReceive();
	}
	// Next activity start (harmonogram)
	stime = currentSlotEnd - ROUTER_TIME();
	if(stime <= CHECK_PHASE_MINIMAL_INTERVAL)
	{
		if(minStep)
			stime = minStep;
		else
			stime = CHECK_PHASE_MINIMAL_INTERVAL;	// uS minimal interval
	}
	newPhaseTimeControl.attach_us(setCurrentPhaseISR, stime);
	toa = ((ROUTER_TIME() - startProc) * 3) / 2;
	if(toa > minStep)
		minStep = toa;
}

// Schema not set or master router not active. Wait for any suitable router
static void syncPhase()
{
	short channel, wint, channelIdx, tt;
	bool special;
	
	if(currentPhase == KissBLE::PROCESS_DISABLED)
		return;
	special = RESERVED_ROUTER;
	if(currentPhase != KissBLE::SYNC_PHASE)
	{	// first entry
		clearGlobals();
		callAppHandler(KissBLE::NODE_RESTARTED);
		if(netParams.idParams.reservePriority == 0 && special)
		{
			kissQueue->call(startAsSpecialNode);
			return;
		}
		currentPhase = KissBLE::SYNC_PHASE;
		// reserve place for lantern frame
		registerFrame(KissBLE::CTRL_LANTERN);
		registerFrame(KissBLE::CTRL_REGISTER);
		LOGGER(logMessage, LOG_LEVEL_MINIMUM, "****** Wait on schema ******\n");
	}
	channelIdx = INIT_CHANNEL_IDX;	// in normal mode depend on syncLoop
	if(channelIdx == KissBLE::SYNC_ROUTER_CHANNEL_IDX && syncLoop > 0)
	{	// after first loop on special (first) channel
		if(special && syncLoop > 2)
		{	// This is special node (planned sync or border) and not special
			// router active. Start this node as special (border/sync) router
			kissQueue->call(startAsSpecialNode);
			return;
		}
		// Normal mode. Continue listenning
	}
	channel = CHANNEL(channelIdx);
	// Wait for router. Shorter listen time in any case (normal or special)
	tt = netParams.advParams.timeSchema.tickTime * KissBLE::TICKS_SCALE;
	wint = tt * ((1 << KissBLE::MAX_TICKS_BITS) + 1);
	tt = 0;
	if(special)
	{
		if(syncLoop == 0)
		{
			tt = radio->receive(channel);
			LOGGER(logMessage, LOG_LEVEL_FRAMES, 
				 "Start listen (special) on channel %d\n", channelIdx);
		}
	}
	else
	{
		tt = radio->receive(channel);
		LOGGER(logMessage, LOG_LEVEL_FRAMES, "Start listen on channel=%d, time=%d\n", 
			 channelIdx, wint);
		if(syncLoop >= KissBLE::MAX_CHANNELS && wint < netParams.advParams.deviceWakeupInterval)
		{
			// Extent listen window. Wait for lantern frame from device node
			wint = netParams.advParams.deviceWakeupInterval;
			LOGGER(logMessage, LOG_LEVEL_FULL, "Extent listen time to %d ms\n", wint);
		}
	}
	if(tt)
	{
		LOGGER(logMessage, LOG_LEVEL_MINIMUM, "***** Start listen error %d\n", tt);
	}
	syncLoop++;
	newPhaseTimeControl.attach_us(syncPhase, wint * 1000);
}

static int64_t syncTime(uint64_t ct, uint64_t receiveTime)
{
	int64_t newOff = 0;
	if(ct)
	{
		newOff = KissTimer::setRouterTime(ct, receiveTime);
		set_time(ROUTER_TIME() / 1000000);	// update RTC clock
	}
	return newOff;
}

// Return EOS flag if no packets for controlled nodes
static bool checkStopListening()
{
	uint16_t nodeMask = 0;
	if(IS_REGISTERED)
	{	// If node (and virtual nodes) not on the list of active nodes, stop listening
		getLocalNode(netParams.idParams.node, &nodeMask);
		virtualNodesMask(&nodeMask);
		if((routerAdvertisement.nodes2wait & nodeMask) == 0 &&
			(routerAdvertisement.extraTimeNodes & nodeMask) == 0)
		{
			radio->stopReceive(true);	// No packets for THIS node in current timeslot
			LOGGER(logStopListen);
			return true;
		}
	}
	return false;
}

static void processAdvertFrame(IFRAME *iframe)
{
	COMMAND_STATE *cs;
	int64_t syncTick;
	int32_t tshift;
	cs = cmdCache.getCellContent(iframe->cmdIdx);
	FRAME *frame = getFrame(cs);
	CTRL_RADVERT_BODY *ainfo = (CTRL_RADVERT_BODY *) frame->payload.body;
	COMM_SCHEMA newSchema = ainfo->routerSchema;
	bool tn = NODE_EQ(masterRouterNode, frame->shead.srcNode);

	if(SCHEMA_ACTIVE && !tn)
		return;
	if(!ainfo->canRegister && !IS_REGISTERED)
	{
		if(SCHEMA_ACTIVE && tn)
			kissQueue->call(syncPhase);	// search next router
		return;
	}
	tshift = TICK_SHIFT_US(newSchema.tickStartShift);
	if(ainfo->routerTime > 0)
	{
		int64_t off = syncTime(ainfo->routerTime, iframe->startTime);
		LOGGER(logRouterTime, ainfo->routerTime, iframe->startTime, off);
		(void) off;
	}
	switch2routerTime(iframe);
	syncTick = (iframe->startTime - tshift);
	activeSchema = ainfo->routerSchema;
	routerAdvertisement = *ainfo;
	tickStartTime = syncTick;
	LOGGER(logAdvertFrame, &frame->shead, syncTick);
	schemaSyncTime = syncTick;
	masterRouterNode = frame->shead.srcNode;
	fillSlotsTable();
	routerInactiveCount = 0;
	// update (shift) timer interrupt call!
	setCurrentPhase(frame);
}

// Currently used only by DEVICE node.
// Not ISR context
static void endOfListenSession(uint64_t frameTime, short line)
{
	radio->stopReceive(true);
	if(currentSlotCompletedTime == 0)
		currentSlotCompletedTime = frameTime;
	LOGGER(endOfListenStat, currentSlotCompletedTime);
	if(ADVERT_ACTIVE)
	{
#if defined(KISS_EMULATOR)
		if(currentSlot == activeSchema.sendTick && currentSlotCompletedTime > 0)
			emulate(currentIntervalNo);
#endif		
		prepare4send();
	}
}

bool KissBLE::transactionEQ(FHEAD *head1, FHEAD *head2)
{
	NODE_ID nd1, nd2;
	bool rv = head1->frameType == head2->frameType && 
			head1->frameSeq == head2->frameSeq;
	if(rv)
	{
		nd1 = (head1->response) ? head1->dstNode : head1->srcNode;
		nd2 = (head2->response) ? head2->dstNode : head2->srcNode;
		rv = CONN_EQ(nd1, nd2);
	}
	return rv;
}

// On DEVICE. On router all received frames are processed
static bool ackFromReceivedFrame(FHEAD *head, short cmd = -1)
{
	COMMAND_STATE *cs = NULL, *csidx = NULL;
	FHEAD *whead;
	short i, vnode, vnode2, fowner = -1, processed = 0;
	bool retVal = false, drop;
	
	fowner = getOwner(head);
	if(cmd >= 0)
		csidx = cmdCache.getCellContent(cmd);
	LOGGER(logStartAck, head);
	vnode = getVirtualNode(head->dstNode);
	vnode2 = getVirtualNode(head->srcNode);
	if(IS_DEVICE && vnode < 0 && vnode2 < 0)
		return false;	// node not involved in transaction
	while((i = cmdCache.nextBusyCell(&cs)) >= 0)
	{
		if(i == cmd)
			continue;
		whead = getHeader(cs);
		if(!KissBLE::transactionEQ(whead, head))
			continue;
		if(processed >= 2)
			break;
		// Response can ACK request frame and vice versa!
		if(whead->response != head->response)
		{	// Response to request
			if(fowner == cs->virtualOwner)
			{
				// ACK for immediate sender (router!)
				if(head->response && head->srcRouter)
				{	// destination DEVICE node? Automatic ACK generated after response.
					if(frameHoops(head) > 1)
					{
						vnode = getVirtualNode(head->dstNode);
						if(vnode >= 0 && csidx && append2ack(head) >= 0)
							csidx->ackSendInt = encodeInterval();
					}
				}
				processed++;
				if(head->response)
					FREE_COMMAND(i, 8);
				else
					retVal = true;
			}
			continue;
		}
		else
		{	// Check duplicate
			if(fowner == cs->virtualOwner)
			{
				if(csidx)
				{
					LOGGER(logIgnoreDuplicate, csidx);
				}
				else
					FREE_COMMAND(i, 9);
				return true;
			}
		}
		// On router ACK only for router owner frames (==0)
		if(!IS_DEVICE)
			if(cs->virtualOwner || head->srcRouter == whead->srcRouter)
				continue;
		// This node or virtual
		processed++;
		if(cs->frameSendInt == 0 || cs->ackReceived)
		{	// not sended or ACK received
			LOGGER(logACKsync);
			continue;
		}
		cs->ackReceived = 1;
		drop = (!cs->wait4response && (whead->srcNode.connPort == 0 || whead->response));
		if(drop)
			FREE_COMMAND(i, 10);
	}
	return retVal;
}

// ISR
static bool reservePriority(FRAME *frame)
{
	bool omit = false;
	FHEAD *head = &frame->shead;
	if(SPECIAL_ROLE(head->srcNode))
	{
		short priority = -1;
		if(head->frameType == KissBLE::CTRL_ADVERTISEMENT)
		{
			CTRL_RADVERT_BODY *adv = (CTRL_RADVERT_BODY *) frame->payload.body;
			priority = adv->reservePriority;
		} else if(head->frameType == KissBLE::CTRL_LANTERN)
		{
			CTRL_LANTERN_BODY *lb = (CTRL_LANTERN_BODY *) frame->payload.body;
			priority = lb->reservePriority;
		}
		if(priority >= 0)
		{
			if(priority > netParams.idParams.reservePriority)
				omit = true;	// ignore advertisement from special routers with lower priority
			else if(priority < netParams.idParams.reservePriority)
			{
				kissQueue->call(syncPhase);
				omit = true;
			}
		}
	}
	return omit;
}

// ISR context
void prefilterPacket(IFRAME *ihead, uint8_t *packet)
{
	short omit = 0, ft, plen = ihead->len, tmp;
	FRAME decodedFrame = {0}, *frame;
	FHEAD *head;
	bool eos, callACK = false, decode = false;
	NODE_STATE *ns;
	uint8_t regFlag;
	uint8_t lctx[KissSecurityManager::WORKSPACE_SIZE] = {0}, *slowContext = NULL;
	
	frame = (FRAME *) packet;
	if(securityManager)
	{
		decode = securityManager->encodedPacket(packet);
		if(decode)
		{
			tmp = KissBLE::MIN_FRAME_LENGTH + ADVERT_BODY_LEN;	// ctrl frames decrypted 
			if(!KissSecurityManager::FAST_MANAGER && plen > tmp)
				slowContext = lctx;	// decode fragment. Otherwise decrypt whole packet
			plen = securityManager->decodeFrame(packet, plen, (uint8_t *)&decodedFrame, slowContext);
			if(plen < 0)
			{
				QUEUE_LOGGER(logDiscardedFrame, 3, plen);
				return;
			}
			else if(plen > 0)
				frame = &decodedFrame;
		}
	}
	// TODO. Serialization
	head = &frame->shead;
	LOGGER(stopOrIgnore, KissLogger::DBG_PREFILTER, (COMMAND_STATE *) head, NULL);
	if(plen < KissBLE::MIN_FRAME_LENGTH || 
		head->frameLen < KissBLE::MIN_FRAME_LENGTH ||
		head->frameLen >= KissBLE::MAX_PACKET_SIZE)
	{
		QUEUE_LOGGER(logDiscardedFrame, 2, head->frameLen);
		return;
	}
	if(netParams.idParams.pan != 0 && head->pan != netParams.idParams.pan)
		omit = __LINE__;
	ft = head->frameType;
	if(!omit)
	{
		if(IS_REGISTERED)
		{
			ns = getLocalNodeState(head->srcNode);
			if(ns)
				ns->nodeActivity = encodeInterval();
		}
		if(thisRouterId)
		{
			if(reservePriority(frame))
				omit = __LINE__;
#if !defined(KISS_EMULATOR)			
			else if(head->srcRouter == thisRouterId)
				omit = __LINE__;
			else if(ft == KissBLE::CTRL_REGISTER && head->response)
				omit = __LINE__;
#endif			
		}
		else
		{	// device or not yet synchronized
			uint8_t myRouter = masterRouterNode.routerId;
			if(myRouter > 0)
			{	// synced
				if(myRouter != head->srcRouter)
					omit = __LINE__;	// wait only for packets from router  (or candidate)
				else
				{
					if(IS_REGISTERED)
					{	// Dest. node is virtual on this device?
						if(!THIS_NODE(head->dstNode) && !EMPTY_NODE(head->dstNode))
						{
							if(getVirtualParam(head->dstNode) < 0)
							{
								omit = __LINE__;
								callACK = true;
							}
						}
					}
					else
					{	// wait only for advertisement or register frame response
						if(ft == KissBLE::CTRL_REGISTER)
						{
							if(head->response)
							{
								CTRL_REGISTER_BODY *rb = (CTRL_REGISTER_BODY *) frame->payload.body;
								if(!KissBLE::macEQ(&netParams.idParams.mac, &rb->mac))
									omit = __LINE__;
							}
							else
								omit = __LINE__;
						}
						else if(ft != KissBLE::CTRL_ADVERTISEMENT)
							omit = __LINE__;	
					}
				}
			} else if(head->frameType != KissBLE::CTRL_ADVERTISEMENT)
				omit = __LINE__;	// wait only for advertisement
		}
	}
	if(!omit && ft == KissBLE::CTRL_LANTERN)
		omit = __LINE__;
	LOGGER(updateStat, omit);
	if(!omit)
	{	// process ctrl frames
		if(decode && securityManager && slowContext && !USER_FRAME(ft))
		{	// decrypt rest of CTRL packet
			short stat = securityManager->decodeFrame(packet, plen, (uint8_t *)&frame, slowContext);
			if(stat < 0)
				omit = __LINE__;
			slowContext = NULL;
		}
		if(!omit && thisRouterId == 0 && ft == KissBLE::CTRL_ADVERTISEMENT)
		{	// monitor EOS from router
			CTRL_RADVERT_BODY *abody = (CTRL_RADVERT_BODY *) frame->payload.body;
			eosFromAdvertisement = abody->eosId;
		}
	}
	eos = checkEOS(head, eosFromAdvertisement);
	if(omit && !callACK && !eos)
	{
		QUEUE_LOGGER(logOmittedFrame, *head, eos, omit);
	}
	else
	{
		regFlag = (omit) ? KissBLE::OMITTED_FRAME : 0;
		regFlag |= KissBLE::WAIT4PROCESS;
		if(eos)
			regFlag |= KissBLE::END_OF_SESSION;
		if(slowContext == NULL)
			head = NULL;
		ihead->cmdIdx = putIntoCache(frame, regFlag, head, ihead->len);
		if(ihead->cmdIdx >= 0)
		{
			framesInQueue++;
			kissQueue->call(processReceivedPacket, *ihead);	// leave ISR context
		}
	}
}

// ISR context
static void receivePacket()
{
	IFRAME ihead = {0};
	uint8_t packet[KissBLE::MAX_PACKET_SIZE];
	short plen = radio->getPacketInfo(&ihead.startTime, &ihead.endTime);
	if(plen <= 0)
		return;	// no packet
	ihead.len = plen = radio->getPacket(&packet, KissBLE::MAX_PACKET_SIZE);
	if(plen < KissBLE::MIN_FRAME_LENGTH)
	{
		LOGGER(logMessage, LOG_LEVEL_FRAMES, "*** Too short frame. Ignored\n");
		return;
	}
	prefilterPacket(&ihead, packet);
}

static void radioHandler(short event)
{
	
	switch(event)
	{
		case Transceiver::RX_OK:
			receivePacket();
			break;
			
		case Transceiver::TX_OK:
			sendNextFrame(0);			// still sending
			break;
			
		case Transceiver::TX_FAIL:
			LOGGER(logMessage, LOG_LEVEL_FRAMES,"Packet sent FAIL\n");
			sendNextFrame(-1);			// still sending
			break;
	}
}

static short isConnected(COMMAND_STATE *cs)
{
	short vn;
	FHEAD *head = getHeader(cs);
	if(THIS_NODE(head->dstNode))
	{	// real node
		NODE_STATE *ns = getLocalNodeState(head->srcNode);
		if(ns)
		{
			if(ns->linkState == KissBLE::INITIATED)
				ns->linkState = KissBLE::CONNECTED;
		}
		if(ns->linkState != KissBLE::CONNECTED)
			return -KissBLE::NODE_NOT_CONNECTED;
	}
	vn = getVirtualNode(head->dstNode);
	if(head->srcNode.connPort || vn > 0)
	{
		CONNECTION_STATE *st = getConnectionState(head->srcNode, vn);
		if(st)
		{
			bool switched = (st->connState == KissBLE::INITIATED);
			if(switched)
				st->connState = KissBLE::CONNECTED;
			if(st->connState == KissBLE::CONNECTED)
			{
				short conn = 0;
				if(head->response == 0 && head->srcNode.connPort)
					conn = checkNextSeq(st, head);
				if(conn)
				{
					LOGGER(logOutOfSeq, cs, st->lastReceivedSeq);
				}
				if(switched && head->srcNode.connPort)
				{
					NODE_ID nd2 = head->srcNode;
					nd2.connPort = 0;
					st = getConnectionState(nd2, head->dstNode.nodeId);
					if(st && st->connState == KissBLE::INITIATED)
						st->connState = KissBLE::CONNECTED;
				}
				if(conn)
					return conn;
			}
			else
				return -KissBLE::NODE_NOT_CONNECTED;
		}
	}
	return 0;
}

CONNECTION_STATE *getConnFromHandle(int32_t connHandle, NODE_ID *destNode = NULL, short *srcNode = NULL)
{
	NODE_ID nd;
	short lnode, connIdx, cstate = 0;
	CONNECTION_STATE *cs = NULL;
	uint16_t *wrk = (uint16_t *) &nd;
	BUILD_BUG_ON(sizeof(NODE_ID) != sizeof(uint16_t));
	*wrk = connHandle & 0xFFFF;
	connHandle >>= 16;
	lnode = connHandle & 0xFF;
	connHandle >>= 8;
	connIdx = (connHandle & 0xFF) - 1;
	if(connIdx >= 0)
	{
		cs = connections.getCellContent(connIdx);
		if(cs)
		{
			short vn = (lnode == netParams.idParams.node.nodeId) ? 0 : lnode;
			if(CONN_EQ(nd, cs->connNode) && cs->virtualOwner == vn)
				cstate = cs->connState;
		}
	}
	else
	{	// real link (transactional) connection
		NODE_STATE *ns = getLocalNodeState(nd);
		if(ns && ns->mac.len)
			cstate = ns->linkState;
	}
	if(cstate != KissBLE::CONNECTED)
	{
		nd = KissBLE::nullNodeId;
		lnode = -KissBLE::NODE_NOT_CONNECTED;
	}
	if(destNode)
		*destNode = nd;
	if(srcNode)
		*srcNode = lnode;
	return cs;
}

int32_t composeConnHandle(CONNECTION_STATE *cs)
{
	int32_t rv;
	NODE_ID connNode = cs->connNode;
	short vowner = cs->virtualOwner;
	uint16_t *wrk = (uint16_t *) &connNode;
	if(vowner == 0)
		vowner = netParams.idParams.node.nodeId;
	rv = (vowner << 16) | *wrk;
	rv |= (connections.getCellId(cs) + 1) << 24;
	return rv;
}

int32_t composeConnHandle(NODE_ID connNode, short vowner = 0)
{
	int32_t rv;
	uint16_t *wrk = (uint16_t *) &connNode;
	BUILD_BUG_ON(sizeof(NODE_ID) != sizeof(uint16_t));
	CONNECTION_STATE *cs = getConnectionState(connNode, vowner);
	if(vowner == 0)
		vowner = netParams.idParams.node.nodeId;
	rv = (vowner << 16) | *wrk;
	if(cs)
		rv |= (connections.getCellId(cs) + 1) << 24;
	return rv;
}

static int32_t composeCmdHandle(short cmdidx)
{
	int32_t rv;
	COMMAND_STATE *cs;
	
	if(cmdidx <= 0)
		return cmdidx;
	if(cmdidx > cmdCache.CACHE_SIZE || !cmdCache.cellIsBusy(cmdidx))
		return -KissBLE::INCORRECT_CMD_HANDLE;
	cs = cmdCache.getCellContent(cmdidx);
	rv = (cs->startInt << 8) | (cmdidx & 0xFF);
	return rv;
}

short getCmdIdxFromHandle(int32_t handle)
{
	COMMAND_STATE *cs;
	uint16_t marker;
	short cmdidx = handle & 0xFF;
	if(cmdidx <= 0 || cmdidx > cmdCache.CACHE_SIZE)
		return -KissBLE::INCORRECT_CMD_HANDLE;
	if(!cmdCache.cellIsBusy(cmdidx))
		return -KissBLE::INCORRECT_CMD_HANDLE;
	cs = cmdCache.getCellContent(cmdidx);
	marker = (handle >> 8);
	if(cs->startInt == marker)
		return cmdidx;
	return -KissBLE::INCORRECT_CMD_HANDLE;
}

static short callAppHandler(short event_type, short cmdIdx)
{
	EVENT event = {0};
	short stat;
	COMMAND_STATE *cs = NULL;
	
	event.eventType = event_type;
	if(cmdIdx >= 0)
	{
		event.cmdHandle = composeCmdHandle(cmdIdx);
		cs = cmdCache.getCellContent(cmdIdx);
	}
	LOGGER(logAppHandler, event_type, cmdIdx);
#if defined(KISS_EMULATOR)	
	if(cs && cs->virtualOwner > 0)
		stat = emulateEventHandler(&event, kissQueue, cs->virtualOwner);
	else
#endif
		stat = appHandler(&event);
	if(cs)
		cs->wait4handler = 0;
	if(stat < 0)
	{
		LOGGER(logMessage, LOG_LEVEL_MINIMUM, "Received packet ignored by handler.\n");
	}
	return stat;
}

static short generateNodeId(MAC_ADDR *dmac, short hint)
{
	short i;
	MAC_ADDR *mac2;
	NODE_STATE *ns = NULL;

	while((i = localNodes.nextBusyCell(&ns)) >= 0)
	{
		mac2 = &ns->mac;
		if(KissBLE::macEQ(dmac, mac2))
			return i;	// node registered in the past
	}
	// No MAC in localNode table
	if(hint > 0 && !localNodes.cellIsBusy(hint))
		ns = localNodes.put(NULL, hint);
	else
		ns = localNodes.put();
	if(ns)
	{
		ns->mac = *dmac;
		i = localNodes.getCellId(ns);
	}
	else
		i = -1;
	return i;
}

static short createConnection(NODE_ID cnode, short vn, short status, NODE_ID destNode)
{
	short st = 0;
	CONNECTION_STATE *connState, ncs = {0};
	ncs.connNode = cnode;
	ncs.virtualOwner = vn;
	connState = connections.put(&ncs);
	if(connState)
	{
		if(cnode.connPort > 0 && vn == 0)
		{
			NODE_STATE *ns = getLocalNodeState(cnode);
			ns->portsInUse |= (0x8 >> cnode.connPort);
		}
		connState->connState = status;
		LOGGER(logCreateConn, cnode, destNode, status);
	}
	else
		st = -KissBLE::NO_FREE_SPACE_IN_CONN_TABLE;
	return st;
}

static short markConnection(FHEAD *head, short port, short status)
{
	short st = 0, vn = getVirtualNode(head->dstNode);
	CONNECTION_STATE *connState;
	NODE_STATE *ns = getLocalNodeState(head->srcNode);
	
	if(status == 0)
		status = (head->response || head->frameType == KissBLE::CTRL_REGISTER) ? 
			KissBLE::CONNECTED : KissBLE::INITIATED;
	if(vn == 0)	// real node.
	{
		if(ns->portsInUse == 0)
		{	// First conn (link)
			ns->linkState = status;
			ns->portsInUse = 0x8;
		}
		LOGGER(logMarkConnection, head, status);
	}
	if(vn > 0 || port > 0)
	{	// Virtual or sequential connection. Mark in table
		NODE_ID cnode = head->srcNode;
		cnode.connPort = port;
		connState = getConnectionState(cnode, vn);
		if(connState == NULL)
			st = createConnection(cnode, vn, status, head->dstNode);
		if(vn > 0 && port > 0)
		{
			cnode.connPort = 0;
			connState = getConnectionState(cnode, vn);
			if(connState == NULL)
				st = createConnection(cnode, vn, status, head->dstNode);
		}
	}
	return st;
}

static void processOpenConnectionFrame(short cmdIdx)
{
	COMMAND_STATE *cs = cmdCache.getCellContent(cmdIdx);
	FHEAD *head = getHeader(cs);
	FRAME *frame = NULL;
	CTRL_CONNECTION_BODY *cb2 = NULL;
	short nd = 0, st, vn;
	NODE_STATE *ns;
	
	if(thisRouterId)
	{
		frame = getFrame(cs);
		cb2 = (CTRL_CONNECTION_BODY *) frame->payload.body;
		ns = getLocalNodeState(head->dstNode);
		if(ns && ns->registered)
		{	// not empty. fill MAC 
			if(cb2->dstMac.len == 0)
			{
				cb2->dstMac = ns->mac;
				cs->changed = 1;
			}
		}
		else if(cb2->dstMac.len)
		{	// fill node id
			vn = getLocalNode(&cb2->dstMac);
			if(vn > 0 && isRegistered(vn))
			{
				head->dstNode = composeNodeId(vn);
				cs->changed = 1;
			}
		}
	}
	vn = frameAtDestination(cs);
	if(vn >= 0)
	{	// register or reject connection
		frame = getFrame(cs);
		cb2 = (CTRL_CONNECTION_BODY *) frame->payload.body;
		st = 0;
		if(head->response == 0)
		{
			st = callAppHandler(KissBLE::CONNECTION_REQUEST, cmdIdx);
			if(st > 0)
				st = 0;
		}
		else
			st = head->netStatus;
		if(st == 0)
		{
			ns = getLocalNodeState(head->srcNode);
			if(ns->mac.len == 0)
				ns->mac = cb2->srcMac;
			st = markConnection(head, cb2->port, head->netStatus);
			LOGGER(logOpenConn, cs, cb2->port, st);
		}
		if(head->response == 0)
		{	// generate response with the same body
			head->netStatus = st;
			pushResponse(cmdIdx);
		}
		else
		{	// response
			callAppHandler(KissBLE::CONNECTION_RESPONSE, cmdIdx);
			FREE_COMMAND(cmdIdx);
		}
	}
	else
	{	// resend to dest node
		if(!head->response)
		{
			frame = getFrame(cs);
			cb2 = (CTRL_CONNECTION_BODY *) frame->payload.body;
			nd = getLocalNode(&cb2->dstMac);
			if(nd > 0)
			{
				if(EMPTY_NODE(head->dstNode))
				{
					head->dstNode = composeNodeId(nd);
					cs->changed = 1;
				}
			}
			else
			{	// node not registered. Negative response
				head->netStatus = KissBLE::NODE_NOT_VISIBLE;
				pushResponse(cmdIdx);
			}
		}
		LOGGER(logResendFrame, cs);
	}
}

static void processRegisterFrame(short cmdIdx)
{
	COMMAND_STATE *cs = cmdCache.getCellContent(cmdIdx);
	FRAME *frame = getFrame(cs);
	CTRL_REGISTER_BODY *cb2 = (CTRL_REGISTER_BODY *) frame->payload.body;
	bool fmac = !KissBLE::macEQ(&cb2->mac, &netParams.idParams.mac);
	NODE_STATE *ns;
	
	if(frame->shead.response == 0)
	{	// request for specific router?
		bool rmac = KissBLE::macEQ(&cb2->routerMAC, &netParams.idParams.mac);
		if(IS_DEVICE || (cb2->routerMAC.len > 0 && !rmac))
			fmac = false;	// Skip request. Not for this router!
		// fill nodeId or MAC in request
		if(fmac && THIS_NODE(frame->shead.dstNode))
		{	// Someone wants to register on this router
			FRAME response;
			short nodeidx = getLocalNode(&cb2->mac);
			if(nodeidx < 0)
				nodeidx = frame->shead.srcNode.nodeId;
			nodeidx = generateNodeId(&cb2->mac, nodeidx);
			memcpy(&response, frame, KissBLE::getLength(frame));
			if(nodeidx > 0)
			{	
				response.shead.srcNode = netParams.idParams.node;
				ns = localNodes.getCellContent(nodeidx);
				ns->nodeActivity = encodeInterval();
				ns->registered = 1;
				response.shead.dstNode = composeNodeId(nodeidx);
				response.shead.response = 1;
				cb2 = (CTRL_REGISTER_BODY *) response.payload.body;
				cb2->routerMAC = netParams.idParams.mac;
				putIntoCache(&response);
				// silent connect to router
				frame->shead.srcNode = response.shead.dstNode;
				markConnection(&frame->shead, 0, 0);
				LOGGER(logRegisterNode, &response.shead, 0);
			}
		}
	}
	else
	{	// This is register frame response. Set new nodeId (real or virtual)
		bool process = (NODE_EQ(masterRouterNode, frame->shead.srcNode) || thisRouterId);
		short enode = getVirtualNode(&cb2->mac);
		if(process && enode >= 0)
		{	// proper router
			bool registerOK = true;
			if(enode == 0)
			{	// real node (first)
				cs = cmdCache.getCellContent(KissBLE::RESERVED_CACHE_CELL_IDX);
				FRAME *lantern = getFrame(cs);
				netParams.idParams.node = frame->shead.dstNode;
				netParams.idParams.pan = frame->shead.pan;
				ns = getLocalNodeState(masterRouterNode);
				if(cb2->routerMAC.len > 0 && ns)
				{
					ns->mac = cb2->routerMAC;
					ns->nodeActivity = encodeInterval();
				}
				// prepare lantern frame for future use
				bzero(lantern, KissBLE::MIN_FRAME_LENGTH);
				lantern->shead.frameType = KissBLE::CTRL_LANTERN;
				composeFrame(lantern);
				// remove register request
				FREE_COMMAND(KissBLE::RESERVED_CACHE_CELL_IDX + 1);
				setScenarioInterval(currentIntervalNo, netParams.idParams.node.nodeId);
			}
			if(registerOK)
			{
				markConnection(&frame->shead, 0, 0);
				ns = getLocalNodeState(frame->shead.dstNode);
				ns->mac = cb2->mac;
				ns->registered = 1;
				ns->nodeActivity = encodeInterval();
				callAppHandler(KissBLE::NODE_REGISTERED, cmdIdx);
			}
			LOGGER(logRegisterNode, &frame->shead, (registerOK) ? 1 : 2);
		}
	}
	FREE_COMMAND(cmdIdx);
}

static short sendNoConnectionResponse(short cmdidx)
{
	PAYLOAD pay = {0};
	CTRL_ERROR_BODY *rb = (CTRL_ERROR_BODY *) pay.body;
	
	pay.bodyType = KissBLE::CTRL_ERROR;
	pay.bodyLen = sizeof(CTRL_ERROR_BODY);
	rb->errorCode = KissBLE::NODE_NOT_CONNECTED;
	LOGGER(logAbortConn, cmdidx);
	cmdidx = pushResponse(cmdidx, &pay);
	return cmdidx;
}

static void execHandler4delayed()
{
	COMMAND_STATE *cs;
	FHEAD *head;
	short vn, cstat, cmdHandler = getSetBit(cmdPending4handler);
	
	while(cmdHandler >= 0)
	{
		cs = cmdCache.getCellContent(cmdHandler);
		cstat = isConnected(cs);
		if(cstat == 0)
		{
			vn = callAppHandler(KissBLE::FRAME_RECEIVED, cmdHandler);
			clearLeftBit(&cmdPending4handler, cmdHandler);
			head = getHeader(cs);
			if(head->noResponse && head->dstNode.connPort && head->response == 0)
				pushResponse(cmdHandler);	// empty
			if(vn)
				FREE_COMMAND(cmdHandler);
			cmdHandler = -1;	// start again from beginnig
		}
		else if(cstat != -KissBLE::FRAME_OUT_OF_SEQUENCE)
		{
			if(cs->wait4response)
			{
				cstat = sendNoConnectionResponse(cmdHandler);
				if(cstat >= 0)
					clearLeftBit(&cmdPending4handler, cmdHandler);
			}
			else
				clearLeftBit(&cmdPending4handler, cmdHandler);
		}
		cmdHandler = getSetBit(cmdPending4handler, cmdHandler + 1);
	}
}

static bool processUserFrame(short cmdidx)
{
	short vn, cstat, ridx;
	COMMAND_STATE *cs = cmdCache.getCellContent(cmdidx);
	FHEAD *head = getHeader(cs);
	bool rv = USER_FRAME(head->frameType);
	if(rv)
	{
		vn = frameAtDestination(cs);
		if(vn >= 0)
		{
			FRAME *frame = getFrame(cs);
			if(frame == NULL)
				return true;
			if(frame->payload.bodyType == KissBLE::CTRL_HEAD_LIST)
			{
				FREE_COMMAND(cmdidx);
				return true;	// special ctrl response frame. Not for user
			}
			head = &frame->shead;
			cstat = isConnected(cs);
			if(cstat == 0)
			{
				vn = callAppHandler(KissBLE::FRAME_RECEIVED, cmdidx);
				if(head->response == 0)
				{	// request
					if(head->noResponse && head->dstNode.connPort)
					{
						ridx = pushResponse(cmdidx); // immediate network internal empty response
						if(ridx >= 0)
							vn = 0;
					}
				}
				else if(frameHoops(head) <= 1)
					cs->ackSendInt = encodeInterval(); // no ACK needed. Remove frame delayed.
				if(vn)
					FREE_COMMAND(cmdidx);
				// check if there are some pending (out of sequence) frames to process
				if(head->srcNode.connPort)
					execHandler4delayed();
			}
			else
			{
				if(cstat == -KissBLE::FRAME_OUT_OF_SEQUENCE)
				{
					cs->wait4handler = 1;
					setLeftBit(&cmdPending4handler, cmdidx);
				}
				else if(cs->wait4response)
					sendNoConnectionResponse(cmdidx);
			}
		}
		else
		{
			LOGGER(logResendFrame, cs);
		}
	}
	return rv;
}

static void processACKframe(FRAME *frame)
{
	HEAD_LIST *ackb = (HEAD_LIST *) frame->payload.body;
	short i, n = ackb->headListLen;
	for(i = 0; i < n; i++)
	{
//		if(frame->shead.frameType == KissBLE::CTRL_ACK)
		ackb->headList[i].srcRouter = frame->shead.srcRouter;
		ackFromReceivedFrame(ackb->headList + i);
	}
}

// NOT in ISR context !!
static void processReceivedPacket(IFRAME iframe)
{
	short ftype, eos, ignore = 0;
	bool remove;
	COMMAND_STATE *cs = cmdCache.getCellContent(iframe.cmdIdx);
	FHEAD *head = getHeader(cs);

	LOGGER(stopOrIgnore, KissLogger::DBG_PROCESS, cs, &ignore);
	if(ignore)
		cs->ignored = 1;
	framesInQueue--;
	LOGGER(logSendReceiveFrame, &iframe, 'R');
	if(ignore)
	{
		FREE_COMMAND(iframe.cmdIdx);
		return;
	}
	eos = cs->endOfSession;
	if(cs->omitted)
	{	// my frame resended by other node? Set ACK?
		if(IS_REGISTERED && !THIS_NODE(head->dstNode) && !EMPTY_NODE(head->dstNode))
		{
			if(getVirtualParam(head->dstNode) < 0)
				ackFromReceivedFrame(head, iframe.cmdIdx);
		}
		goto end_proc1;
	}
	ftype = head->frameType;
	LOGGER(stopTestSessions, 'R', cs, NULL);
	if(ftype == KissBLE::CTRL_ADVERTISEMENT)
	{	// Frame body decoded in prefilter packet module
		processAdvertFrame(&iframe);
		if(checkStopListening())
			eos = 1;
		goto end_proc1;
	}
	if(!IS_REGISTERED)
	{	// process only ADV and REGISTER response. After process, frame is removed
		if(ftype == KissBLE::CTRL_REGISTER)
		{
			if(SCHEMA_ACTIVE && head->response)
			{
				processRegisterFrame(iframe.cmdIdx);
				goto end_proc;
			}
		}
		goto end_proc1;
	}
	if(ftype == KissBLE::CTRL_ACK)
	{
		FRAME *frame = getFrame(cs);
		if(frame)
			processACKframe(frame);
		goto end_proc1;
	} else if(USER_FRAME(ftype) && head->response && head->userMultiResponse)
	{
		FRAME *frame = getFrame(cs);
		if(frame->payload.bodyLen > 0 && frame->payload.bodyType == KissBLE::CTRL_HEAD_LIST)
			processACKframe(frame);
	}
	// Normal state. Node registered.
	remove = ackFromReceivedFrame(head, iframe.cmdIdx);
	if(remove)
		goto end_proc1;
	LOGGER(incrementFrames);
	if(processUserFrame(iframe.cmdIdx))
		goto end_proc;
	// Special frame
	switch(ftype)
	{
		case KissBLE::CTRL_REGISTER:
			processRegisterFrame(iframe.cmdIdx);
			break;

		case KissBLE::CTRL_OPEN_CONNECTION:
			processOpenConnectionFrame(iframe.cmdIdx);
			break;
	}
	goto end_proc;
end_proc1:
	// Remove received frame
	FREE_COMMAND(iframe.cmdIdx);
	cs = NULL;
end_proc:
	// In any case EOS detected in frame trigger end of session action
	if(cs)
		cs->endOfSession = cs->wait4process = 0;
	if(SCHEMA_ACTIVE && eos)
	{
		switch2routerTime(&iframe);
		uint64_t eosTime = iframe.endTime;
		endOfListenSession(eosTime, __LINE__);
	}
}

static short checkAdvParams(ADVERT_PARAMS *nParams)
{
	return 0;
}

static void startJob()
{
#if defined(APP_LOG)
	LOGGER(logMessage, LOG_LEVEL_MINIMUM, "(Re)start node\n");
	LOGGER(switchLogLevel, true);
#endif
	currentPhase = KissBLE::IDLE_PHASE;
	syncPhase();
}

//----------------------- Class implementation -------------------------------
void KissBLE::setEventHandler(short (*handler)(EVENT *event))
{
	appHandler = handler;
}

void KissBLE::setSecurityManager(KissSecurityManager *manager, short method)
{
	securityManager = manager;
	securityMethod = method;
}

PROCESS_PHASE KissBLE::doJob(short mode)
{
	if(mode > 0 && currentPhase == KissBLE::PROCESS_DISABLED)
	{	// start
		radio->setEventHandler(radioHandler);
		startJob();
	}
	if(mode < 0 && currentPhase != KissBLE::PROCESS_DISABLED)
		stopJob();
	return currentPhase;
}

void KissBLE::getNetworkParams(NETWORK_PARAMS *buf)
{
	*buf = netParams;
}

short KissBLE::updateParams(NETWORK_PARAMS *buf)
{
	short rv = checkAdvParams(&buf->advParams);
	if(rv == 0)
		netParams = *buf;
	return rv;
}

int32_t KissBLE::sendTransaction(int32_t connHandle, PAYLOAD *body, FRAME_FLAGS flags)
{
	short st, cmd, owner;
	FRAME frame;
	NODE_ID destNode;
	int32_t cmdHandle;
	
	getConnFromHandle(connHandle, &destNode, &owner);
	if(owner < 0)
		return 0;
	frame.shead.frameType = body->bodyType;
	st = composeFrame(&frame);
	if(st)
		return 0;
	frame.shead.dstNode = destNode;
	frame.shead.emulated = flags.emulated;
	frame.shead.noResponse = flags.noResponse;
	st = setBody(&frame, body);
	if(st < 0)
		return st;
	if(owner > 0)
	{
		frame.shead.srcNode = composeNodeId(owner);
		frame.shead.srcNode.connPort = frame.shead.dstNode.connPort;
	}
	cmd = putIntoCache(&frame);
	if(cmd < 0)
		return cmd;
	cmdHandle = composeCmdHandle(cmd);
	return cmdHandle;
}

int32_t KissBLE::sendResponse(int32_t cmdHandle, PAYLOAD *body)
{
	short respidx, reqidx = getCmdIdxFromHandle(cmdHandle);
	int32_t rv = 0;

	if(reqidx >= 0)
	{
		COMMAND_STATE *cs = getCommandState(reqidx);
		bool w4r = cs->wait4response;
		respidx = pushResponse(reqidx, body);
		if(respidx >= 0 && !w4r)
			return -KissBLE::RESPONSE_NOT_ALLOWED;
		rv = composeCmdHandle(respidx);
	}
	return rv;
}

short KissBLE::getSrcFrame(int32_t cmdHandle, FRAME **frame)
{
	short cmdidx = getCmdIdxFromHandle(cmdHandle);
	if(cmdidx >= 0)
	{
		COMMAND_STATE *cs = cmdCache.getCellContent(cmdidx);
		*frame = getFrame(cs);
		cmdidx = 0;
	}
	else
		*frame = NULL;
	return cmdidx;
}

static short sendBytesStream(CONNECTION_STATE *seq, const void *bytes, short bytesLen)
{
	FRAME frame;
	short rv = 0, st, len, cmdidx;
	uint8_t *buf = (uint8_t *) bytes;
	frame.shead.frameType = KissBLE::USER_STREAM_BYTE;
	st = composeFrame(&frame);
	if(st < 0)
		return 0;
	frame.shead.dstNode = seq->connNode;
	frame.shead.srcNode.connPort = seq->connNode.connPort;
	if(seq->virtualOwner)
		frame.shead.srcNode.nodeId = seq->virtualOwner;
	frame.shead.noResponse = 1;
	while(bytesLen > 0)
	{
		setBody(&frame);	// clear payload
		len = (bytesLen > KissBLE::MAX_BODY_SIZE) ? KissBLE::MAX_BODY_SIZE : bytesLen;
		st = append2body(&frame, len, buf);
		if(st < 0)
			break;
		cmdidx = putIntoCache(&frame);
		if(cmdidx < 0)
			break;
		// registered for send
//		seq->nextSendSeq++;
		rv += len;
		bytesLen -= len;
		buf += len;
	}
	return rv;
}

short KissBLE::sendBytes(int32_t connHandle, const void *bytes, short bytes2send)
{
	short rv;
	CONNECTION_STATE *cs = getConnFromHandle(connHandle);
	if(cs == NULL)
		return -INCORRECT_CONN_HANDLE;
	rv = sendBytesStream(cs, bytes, bytes2send);
	return rv;
}

static short getFreePort(int32_t connHandle)
{
	NODE_ID dest;
	short i, owner, port = -KissBLE::NO_FREE_STREAM_PORT, shift;
	uint32_t mask = 0;
	CONNECTION_STATE *cs = getConnFromHandle(connHandle, &dest, &owner);
	if(owner == 0 || owner == netParams.idParams.node.nodeId)
	{
		NODE_STATE *ns = getLocalNodeState(dest);
		if(!ns || ns->mac.len == 0)
			return -KissBLE::NO_FREE_STREAM_PORT;
		mask = ns->portsInUse;
	}
	else
	{
		cs = NULL;
		if(owner == netParams.idParams.node.nodeId)
			owner = 0;
		while((i = connections.nextBusyCell(&cs)) >= 0)
		{
			if(NODE_EQ(cs->connNode, dest) && cs->virtualOwner == owner)
				mask |= (0x8 >> cs->connNode.connPort);
		}
	}
	shift = 32 - (1 << KissBLE::CONNECTION_ID_BITS);
	mask <<= shift;
	i = getUnsetBit(mask);
	if(i < (1 << KissBLE::CONNECTION_ID_BITS))
		port = i;
	return port;
}

short KissBLE::getConnectionPort(int32_t connHandle)
{
	short rv;
	NODE_ID dest;
	
	getConnFromHandle(connHandle, &dest, &rv);
	if(rv >= 0)
		rv = dest.connPort;
	return rv;
}

int32_t KissBLE::getConnHandle(MAC_ADDR *destMAC, short port)
{
	int32_t handle;
	short node, i;
	NODE_ID dnode;
	NODE_STATE *ns;
	CONNECTION_STATE *cs;
			
	if(destMAC->len == 0)
		return -MAC_ADDR_UNKNOWN;
	node = getLocalNode(destMAC);
	if(node <= 0)
	{	// check in active command
		COMMAND_STATE *cs = NULL;
		CTRL_CONNECTION_BODY *cb;
		FHEAD *head;
		while((i = cmdCache.nextBusyCell(&cs)) >= 0)
		{
			head = getHeader(cs);
			if(head->frameType == CTRL_OPEN_CONNECTION && cs->virtualOwner == 0 
				&& head->response == 0)
			{
				FRAME *frame = ::getFrame(cs);
				if(frame)
				{
					cb = (CTRL_CONNECTION_BODY *) frame->payload.body;
					if(macEQ(&cb->dstMac, destMAC) && cb->port == port)
						return -INITIATED;
				}
			}
		}
		return -MAC_ADDR_UNKNOWN;
	}
	dnode = composeNodeId(node);
	dnode.connPort = port;
	ns = getLocalNodeState(dnode);
	if(port)
	{
		cs = getConnectionState(dnode);
		if(cs && (ns->portsInUse & (0x8 >> port)))
		{
			if(cs->connState != CONNECTED)
				return -cs->connState;
			handle = composeConnHandle(dnode);
		}
		else
			return -CONN_NOT_INITIATED;
	}
	else
	{
		if(ns->linkState == CONNECTED)
			handle = composeConnHandle(dnode);
		else
			return -ns->linkState;
	}
	return handle;
}

// Open only from THIS node.
int32_t KissBLE::openLink(MAC_ADDR *destMAC, short port)
{
	short i;
	int32_t handle = 0;
	NODE_ID destNode;
	NODE_STATE *ns;
	
	if(destMAC->len == 0)
		return -MAC_ADDR_UNKNOWN;
	i = getLocalNode(destMAC);
	if(i >= 0)
	{
		ns = localNodes.getCellContent(i);
		if(ns->linkState == CONNECTED)
		{
			destNode = composeNodeId(i);
			handle = composeConnHandle(destNode);
		}
	}
	if(handle && port)
		handle = openStream(handle, port);
	else if(handle == 0)
	{
		FRAME frame = {0};
		frame.shead.frameType = CTRL_OPEN_CONNECTION;
		i = composeFrame(&frame, destMAC);
		if(i == 0)
		{
			if(port)
			{
				CTRL_CONNECTION_BODY *cb = (CTRL_CONNECTION_BODY *) frame.payload.body;
				cb->port = port;
			}
			i = putIntoCache(&frame);
			if(i < 0)
				handle = i;	// error
		}
	}
	return handle;
}

int32_t KissBLE::openStream(int32_t linkHandle, short port)
{
	int32_t rv = -1;
	short i, owner;
	NODE_ID dest;
	NODE_STATE *ns;
	FRAME frame = {0};
	getConnFromHandle(linkHandle, &dest, &owner);
	if(owner == 0)
		owner = netParams.idParams.node.nodeId;
	ns = localNodes.getCellContent(owner);
	if(ns == NULL || !isRegistered(owner))
		return -KissBLE::INCORRECT_CONN_HANDLE;
	if(port <= 0)
	{
		port = getFreePort(linkHandle);
		if(port < 0)
			return port;
	}
	else
	{	// Check if busy
		NODE_ID dest2 = dest;
		dest2.connPort = port;
		CONNECTION_STATE *cs = getConnectionState(dest2, owner);
		if(cs && (cs->connState == KissBLE::CONNECTED || cs->connState == KissBLE::INITIATED))
			return composeConnHandle(dest2, owner);
	}
	frame.shead.frameType = CTRL_OPEN_CONNECTION;
	i = composeFrame(&frame, &ns->mac);
	if(i == 0)
	{
		if(port > 0)
		{
			CTRL_CONNECTION_BODY *cb = (CTRL_CONNECTION_BODY *) frame.payload.body;
			cb->port = port;
			dest.connPort = 0;
			frame.shead.dstNode = dest;
			frame.shead.srcNode = composeNodeId(owner);
			cb->srcMac = ns->mac;
			ns = getLocalNodeState(dest);
			cb->dstMac = ns->mac;
			i = putIntoCache(&frame);
			rv = (i < 0) ? i : 0;
		}
		else
			rv = port;
	}
	return rv;
}

static void removeConnectionFrames(NODE_ID destNode)
{
	COMMAND_STATE *cs = NULL;
	FHEAD *head;
	short i;
	
	while((i = cmdCache.nextBusyCell(&cs)) >= 0)
	{
		head = getHeader(cs);
		if(USER_FRAME(head->frameType))
		{
			if(NODE_EQ(destNode, head->dstNode) || NODE_EQ(destNode, head->srcNode))
				FREE_COMMAND(i);
		}
	}
}

short KissBLE::closeConnection(int32_t connHandle, bool immediately)
{
	NODE_ID dest;
	CONNECTION_STATE *seq = getConnFromHandle(connHandle, &dest);
	// clear connection data
	if(seq)
	{
		seq->connState = KissBLE::CLOSE_IN_PROGRESS;
		if(immediately)
		{
			removeConnectionFrames(seq->connNode);
			seq->connState = 0;
		}
	}	
	return 0;
}

short KissBLE::init(Transceiver *rd, short (*random)(void *buf, uint8_t bufLen))
{
	short psize, mchannel, rv;
	BUILD_BUG_ON(sizeof(FRAME) != KissBLE::MAX_PACKET_SIZE);
	bzero(&netParams, sizeof(netParams));
	radio = rd;
	rv = radio->init();
	mchannel = radio->getMaxChannel(&psize);
	if(psize < MAX_PACKET_SIZE || mchannel < MAX_CHANNELS)
		return -1;
	randomData = random;
	radioInterruptNo = (IRQn_Type) radio->getInterruptNo();
	return rv;
}
//---------------------------------------------------------------------

short initKissBLE(Transceiver *trv, EventQueue *globalQueue, short (*random)(void *buf, uint8_t bufLen))
{
	short rv;
	kissQueue = globalQueue;
	rv = KissBLE::init(trv, random);
	if(rv >= 0)
		netDefaultParams();
	return rv;
}
