#pragma once
#include <stdint.h>
#include "Transceiver.h"
#include "KissSecurityManager.h"

class KissSecurityManager;

class KissBLE
{
public:
	
	static const uint8_t	TICKS_SCALE				= 5;	// tick time multiplier
	static const uint8_t	CHANNELS_IDX_BITS		= 2;
	static const uint8_t	MAX_CHANNELS			= ((1 << CHANNELS_IDX_BITS) - 1);
	static const uint8_t	MAX_PACKET_SIZE			= 64;
	static const uint8_t	MAX_LOCAL_NODES			= 16;
	static const uint8_t	MAX_NET_ROUTERS			= 16;
	static const uint8_t	MAX_ADV_CHANNELS		= MAX_CHANNELS;
	static const int16_t	channels[MAX_CHANNELS + 1];
	static const uint8_t	SHIFT_SCALE				= 15;	// in uS
	static const uint8_t	NODE_ID_BITS			= 4;
	static const uint8_t	ROUTER_ID_BITS			= 3;
	static const uint8_t	CONNECTION_ID_BITS		= 2;
	static const uint8_t	MAX_NODE_ID_BITS		= 16;
	static const uint8_t	FRAME_TYPE_BITS			= 4;
	static const uint8_t	CSTATUS_BITS			= 3;
	static const uint8_t	MAX_TICKS_BITS			= 3;
	static const uint8_t	SEND_QUEUE_SIZE			= 16;
	static const uint8_t	NODE_START_ID			= 2;
	static const uint8_t	ROUTER_NODEID			= 1;
	static const uint8_t	SYNC_ROUTER_NET_SLOT	= 1;
	static const uint8_t	SYNC_ROUTER_CHANNEL_IDX	= 1;
	static const uint8_t	CACHE_RESERVED_CELLS	= 1;
	static const uint8_t	RESERVED_CACHE_CELL_IDX	= 0;
	static const uint8_t	REGISTER_GAP_IDX		= 1;
	static const uint8_t	CTRL_FRAME_TYPE_BASE	= 4;
	static const uint8_t	CMD_CACHE_SIZE			= 24;
	static const uint8_t	MAX_CONNECTIONS			= 20;
	static const uint8_t	MAX_FRAMES_PER_STREAM	= 5;
	static const uint8_t	TIME_TO_LIVE			= 6;	// in intervals
	static const uint8_t	WAIT4ACK_TIME			= 2;	// in intervals
	
	enum {
		DEVICE			= 4,
		ROUTER			= 3,
		SYNC_ROUTER		= 2,
		BORDER_ROUTER	= 1,
		ROLE_ID_BITS	= 3,
		ROLE_ID_MASK	= ((1 << ROLE_ID_BITS) - 1)
	} ROLE;
	
	typedef enum {
		TRANSACTIONAL			= 0,
		SEQUENTIAL				= 1,
		CONNECTION_MODE_MASK	= 1
	} CONNECTION_MODE;

	enum {
		NODE_RESTARTED		= 1,
		NODE_REGISTERED		= 2,
		CONNECTION_REQUEST	= 3,
		CONNECTION_RESPONSE	= 4,
		FRAME_RECEIVED		= 5,
		CONNECTION_CLOSED	= 6,
		EVENT_TYPE_MASK		= 0x7
	} EVENT_TYPE;

	typedef enum {
		INITIATED			= 1,
		CONNECTED			= 2,
		FORBIDDEN			= 3,
		NODE_NOT_VISIBLE	= 4,
		OUT_OF_RESOURCES	= 5,
		CLOSE_IN_PROGRESS	= 6,
		CONN_NOT_INITIATED	= 7,
		CONN_STATUS_MASK	= 7
	} CONN_STATUS;

	typedef enum {
		NODE_NOT_CONNECTED			= 1,
		NODE_NOT_ATTAINABLE			= 2,
		PATH_TO_NODE_BROKEN			= 3,
		CONNECTION_REJECTED			= 4,
		MAC_ADDR_UNKNOWN			= 5,
		INCORRECT_CMD_HANDLE		= 6,
		NO_FREE_SPACE_IN_CACHE		= 7,
		INCORRECT_CONN_HANDLE		= 8,
		CONNECTION_PORT_MODE_ERROR	= 9,
		NO_FREE_SPACE_IN_CONN_TABLE	= 10,
		NO_FREE_STREAM_PORT			= 11,
		NO_FRAME_DATA				= 12,
		TOO_MANY_FRAMES_PER_STREAM	= 13,
		FRAME_OUT_OF_SEQUENCE		= 14,
		RESPONSE_NOT_ALLOWED		= 15,
		FRAME_SECURITY_BROKEN		= 16,
		NET_ERRORS_MASK				= 0xFF
	} KISS_ERRORS;

	typedef struct {
		uint8_t noResponse:1;
		uint8_t emulated:1;
		uint8_t dummy: (8 - 2);
	} __attribute__((packed)) FRAME_FLAGS;
	
	typedef struct
	{
		uint16_t	routerId:ROUTER_ID_BITS;
		uint16_t	nodeId	:NODE_ID_BITS;
		uint16_t	role	:ROLE_ID_BITS;
		uint16_t	connPort:CONNECTION_ID_BITS;
		uint16_t	dummy	:(16 - NODE_ID_BITS - ROUTER_ID_BITS - ROLE_ID_BITS - CONNECTION_ID_BITS);
	} __attribute__((packed)) NODE_ID;

	typedef struct
	{
		uint8_t		tickTime:5;						// in TICK_SCALE * (tickTime + 1) ms
		uint8_t		maxTickCount:MAX_TICKS_BITS;	// interval == (maxTickCount + 1) * tickTime
		uint16_t	sendTick:MAX_TICKS_BITS;		// only 1 transmition tick
		uint16_t	sendChannel:CHANNELS_IDX_BITS;	// channel assigned to router
		uint16_t	deviceTick:MAX_TICKS_BITS;
		uint16_t	deviceChannel:CHANNELS_IDX_BITS;
		uint16_t	dummy:(16 - 2*MAX_TICKS_BITS - 2*CHANNELS_IDX_BITS);
		uint16_t	tickStartShift;					// in SHIFT_SCALE uSec
	} __attribute__((packed)) COMM_SCHEMA;

	typedef struct
	{
		uint16_t		deviceWakeupInterval;	// in intervals
		COMM_SCHEMA		timeSchema;
	} __attribute__((packed)) ADVERT_PARAMS;

// hide low-level  technical details
#include "KissPacket.h"	
	
	typedef struct 
	{
		NODE_ID		connNode;
		uint8_t		lastReceivedSeq;
		uint8_t		nextSendSeq;
		uint8_t		dummy:1;
		uint8_t		connState:3;
		uint8_t		virtualOwner:NODE_ID_BITS;
	} __attribute__((packed)) CONNECTION_STATE;

	typedef struct 
	{
		uint16_t	nodeActivity;
		MAC_ADDR	mac;
		uint8_t		portsInUse:4;		// bitmap
		uint8_t		linkState:3;
		uint8_t		registered:1;
	} __attribute__((packed)) NODE_STATE;
	
	typedef struct
	{
		MAC_ADDR		mac;			// MAC address (device identity)
		NODE_ID			node;
		uint8_t			reserveRole;	// border/sync/normal/device. Can be ZERO	
		uint8_t			reservePriority;
		uint8_t			pan;			// PAN id
	} __attribute__((packed)) IDENTITY_PARAMS;
	
	typedef struct
	{
		ADVERT_PARAMS	advParams;
		IDENTITY_PARAMS	idParams;
		uint16_t		transmitDelay;	// in uS
		uint16_t		minTransmitDelay;
	} __attribute__((packed)) NETWORK_PARAMS;
	
	typedef struct
	{
		short			eventType;
		int32_t			cmdHandle;
	} EVENT;

	static short getLength(FRAME *frame) {
		return (frame->shead.frameLen);
	};
	static short sendBytes(int32_t connHandle, const void *bytes, short bytes2send);
	static int32_t sendResponse(int32_t cmdHandle, PAYLOAD *newBody = NULL);
	static int32_t sendTransaction(int32_t connHandle, PAYLOAD *body, FRAME_FLAGS flags);
	static short getSrcFrame(int32_t cmdHandle, FRAME **frame);
	static int32_t getConnHandle(int32_t cmdHandle);
	static int32_t openLink(MAC_ADDR *destMAC, short port = 0);
	static int32_t openStream(int32_t linkHandle, short port = 0);
	static int32_t getConnHandle(MAC_ADDR *destMAC, short port = 0);
	static short closeConnection(int32_t connHandle, bool immediately = true);
	static void getNetworkParams(NETWORK_PARAMS *buf);
	static short updateParams(NETWORK_PARAMS *buf);
	static short getConnectionPort(int32_t connHandle);
	static bool macEQ(MAC_ADDR *mac1, MAC_ADDR *mac2) {
		return (mac1->len == mac2->len && memcmp(mac1->bytes, mac2->bytes, mac1->len) == 0);
	};
	
	static bool transactionEQ(FHEAD *head1, FHEAD *head2);
	static inline bool nodeIdNull(NODE_ID id) {
		return id.nodeId == 0;
	};
	
	static inline uint16_t decodeInterval(uint16_t encodedInt) {
		return encodedInt >> 1;
	};
	static short init(Transceiver *radio, short (*random)(void *buf, uint8_t bufLen));
	static void setEventHandler(short (*handler)(EVENT *));
	static void setSecurityManager(KissSecurityManager *manager, short method = 0);
	static PROCESS_PHASE doJob(short mode = 0);
	
private:
#include "disableNew.h"
};

short initKissBLE(Transceiver *trv, EventQueue *globalQueue, short (*random)(void *buf, uint8_t bufLen));
